from __future__ import annotations

import abc
import functools
import typing as t
from http import HTTPStatus

import attrs
import httpx

from _bentoml_sdk import IODescriptor
from bentoml.exceptions import BentoMLException

T = t.TypeVar("T")


def map_exception(resp: httpx.Response) -> BentoMLException:
    status = HTTPStatus(resp.status_code)
    exc = BentoMLException.error_mapping.get(status, BentoMLException)
    return exc(resp.text, error_code=status)


@attrs.define(slots=True)
class ClientEndpoint:
    name: str
    route: str
    doc: str | None = None
    input: dict[str, t.Any] = attrs.field(factory=dict)
    output: dict[str, t.Any] = attrs.field(factory=dict)
    input_spec: type[IODescriptor] | None = None
    output_spec: type[IODescriptor] | None = None
    stream_output: bool = False
    is_task: bool = False


class AbstractClient(abc.ABC):
    endpoints: dict[str, ClientEndpoint]

    def __init__(self) -> None:
        for name in self.endpoints:
            if name == "__call__":
                # __call__ must be set on the class
                continue
            attr_name = name
            if getattr(self, attr_name, None) is not None:
                attr_name = f"api_{name}"  # prefix to avoid name conflict
            setattr(self, attr_name, self._make_method(name))

    def _make_method(self, name: str) -> t.Callable[..., t.Any]:
        endpoint = self.endpoints[name]

        def method(*args: t.Any, **kwargs: t.Any) -> t.Any:
            return self.call(name, *args, **kwargs)

        method.__doc__ = endpoint.doc
        if endpoint.input_spec is not None:
            method.__annotations__ = endpoint.input_spec.__annotations__
            method.__signature__ = endpoint.input_spec.__signature__
        if endpoint.is_task:
            method.submit = functools.partial(self._submit, endpoint)
            method.get = functools.partial(self._get_task_result, endpoint)
            method.get_status = functools.partial(self._get_task_status, endpoint)
            method.cancel = functools.partial(self._cancel_task, endpoint)
            method.retry = functools.partial(self._retry_task, endpoint)
        return method

    @abc.abstractmethod
    def call(self, __name: str, /, *args: t.Any, **kwargs: t.Any) -> t.Any:
        """Call a service method by its name.
        It takes the same arguments as the service method.
        """

    @abc.abstractmethod
    def _submit(
        self, __endpoint: ClientEndpoint, /, *args: t.Any, **kwargs: t.Any
    ) -> t.Any:
        """Submit a job to the service.
        It takes the same arguments as the service method.
        """

    def _get_task_status(self, __endpoint: ClientEndpoint, /, task_id: str) -> t.Any:
        """Get the status of a task."""

    def _cancel_task(self, __endpoint: ClientEndpoint, /, task_id: str) -> t.Any:
        """Cancel a task."""

    def _get_task_result(self, __endpoint: ClientEndpoint, /, task_id: str) -> t.Any:
        """Get the result of a task."""

    def _retry_task(self, __endpoint: ClientEndpoint, /, task_id: str) -> t.Any:
        """Retry a task."""

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        if "__call__" not in self.endpoints:
            raise TypeError("This service is not callable.")
        return self.call("__call__", *args, **kwargs)



from __future__ import annotations

import hashlib
import json
import os
import shutil
import typing as t
from functools import cached_property

import attrs
from fs.base import FS

from bentoml._internal.bento.bento import BentoModelInfo
from bentoml._internal.cloud.schemas.modelschemas import ModelManifestSchema
from bentoml._internal.cloud.schemas.schemasv1 import CreateModelSchema
from bentoml._internal.models.model import ModelContext
from bentoml._internal.tag import Tag
from bentoml._internal.types import PathType

from .base import Model

if t.TYPE_CHECKING:
    from huggingface_hub import HfApi

CONFIG_FILE = "config.json"
DEFAULT_HF_ENDPOINT = "https://huggingface.co"


@attrs.define(unsafe_hash=True)
class HuggingFaceModel(Model[str]):
    """A model reference to a Hugging Face model.

    Args:
        model_id (str): The model tag. E.g. "google-bert/bert-base-uncased".
        revision (str, optional): The revision to use. Defaults to "main".
        endpoint (str, optional): The Hugging Face endpoint to use. Defaults to https://huggingface.co.
        include (List[str], optional): The files to include. Defaults to all files.
        exclude (List[str], optional): The files to exclude. Defaults to no files.

    Returns:
        str: The downloaded model path.
    """

    model_id: str
    revision: str = "main"
    endpoint: t.Optional[str] = attrs.field(factory=lambda: os.getenv("HF_ENDPOINT"))
    include: t.Optional[t.List[str]] = None
    exclude: t.Optional[t.List[str]] = None

    @cached_property
    def _hf_api(self) -> HfApi:
        from huggingface_hub import HfApi

        return HfApi(endpoint=self.endpoint)

    @cached_property
    def commit_hash(self) -> str:
        return (
            self._hf_api.model_info(self.model_id, revision=self.revision).sha
            or self.revision
        )

    def resolve(self, base_path: t.Union[PathType, FS, None] = None) -> str:
        from huggingface_hub import snapshot_download

        if isinstance(base_path, FS):
            base_path = base_path.getsyspath("/")

        snapshot_path = snapshot_download(
            self.model_id,
            revision=self.revision,
            endpoint=self.endpoint,
            cache_dir=os.getenv("BENTOML_HF_CACHE_DIR"),
            allow_patterns=self.include,
            ignore_patterns=self.exclude,
        )
        if base_path is not None:
            model_path = os.path.dirname(os.path.dirname(snapshot_path))
            os.makedirs(base_path, exist_ok=True)
            shutil.copytree(
                model_path,
                os.path.join(base_path, os.path.basename(model_path)),
                symlinks=True,
                dirs_exist_ok=True,
            )
        return snapshot_path

    def to_info(self, alias: str | None = None) -> BentoModelInfo:
        model_id = self.model_id.lower()
        metadata = {
            "model_id": model_id,
            "revision": self.commit_hash,
            "endpoint": self.endpoint or DEFAULT_HF_ENDPOINT,
            "include": self.include,
            "exclude": self.exclude,
        }
        content_hash = hashlib.md5(
            json.dumps(metadata, sort_keys=True, separators=(",", ":")).encode()
        ).hexdigest()
        tag = Tag(model_id.replace("/", "--"), content_hash)
        return BentoModelInfo(
            tag, alias=alias, registry="huggingface", metadata=metadata
        )

    @classmethod
    def from_info(cls, info: BentoModelInfo) -> HuggingFaceModel:
        if not info.metadata:
            name = info.tag.name
            return cls(model_id=name.replace("--", "/"))
        model = cls(
            model_id=info.metadata["model_id"],
            revision=info.metadata["revision"],
            endpoint=info.metadata["endpoint"],
            include=info.metadata.get("include"),
            exclude=info.metadata.get("exclude"),
        )
        # the commit hash is frozen in the model info, update the cache directly
        model.__dict__.update(commit_hash=info.metadata["revision"])
        return model

    def _get_model_size(self, revision: str) -> int:
        from huggingface_hub.utils import filter_repo_objects

        info = self._hf_api.model_info(
            self.model_id, revision=revision, files_metadata=True
        )
        filtered_files = filter_repo_objects(
            items=info.siblings or [],
            allow_patterns=self.include,
            ignore_patterns=self.exclude,
            key=lambda f: f.rfilename,
        )
        return sum((file.size or 0) for file in filtered_files)

    def to_create_schema(self) -> CreateModelSchema:
        context = ModelContext(framework_name="huggingface", framework_versions={})
        endpoint = self.endpoint or DEFAULT_HF_ENDPOINT
        revision = self.commit_hash
        url = f"{endpoint}/{self.model_id}/tree/{revision}"
        metadata = {
            "registry": "huggingface",
            "model_id": self.model_id,
            "revision": revision,
            "endpoint": endpoint,
            "include": self.include,
            "exclude": self.exclude,
            "url": url,
        }
        return CreateModelSchema(
            description="",
            version=self.to_info().tag.version or revision,
            manifest=ModelManifestSchema(
                module="",
                metadata=metadata,
                api_version="v1",
                bentoml_version=context.bentoml_version,
                size_bytes=self._get_model_size(revision),
                context=context.to_dict(),
                options={},
            ),
        )

    def __str__(self) -> str:
        return f"{self.model_id}:{self.revision}"



from __future__ import annotations

import logging
import os
import sys
import tarfile
from io import BytesIO
from pathlib import Path

import fs

from ...exceptions import BentoMLException
from ...grpc.utils import LATEST_PROTOCOL_VERSION
from ..configuration import is_editable_bentoml
from ..utils.pkg import source_locations

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

logger = logging.getLogger(__name__)

BENTOML_DEV_BUILD = "BENTOML_BUNDLE_LOCAL_BUILD"


def build_bentoml_sdist(
    target_path: str, *, _internal_protocol_version: str = LATEST_PROTOCOL_VERSION
) -> str | None:
    """
    This is for BentoML developers to create Bentos that contains the local bentoml
    build based on their development branch. To disable this behavior, one should
    set envar :code:`BENTOML_BUNDLE_LOCAL_BUILD=false` before building a Bento.

    Returns the sdist filename if the build is successful, otherwise None.
    """
    import tomli_w

    from ..configuration import BENTOML_VERSION

    if os.environ.get(BENTOML_DEV_BUILD, "true").lower() == "false":
        return

    if not is_editable_bentoml():
        # skip this entirely if BentoML is installed from PyPI
        return

    # Find bentoml module path
    # This will be $GIT_ROOT/src/bentoml
    module_location = source_locations("bentoml")
    if not module_location:
        raise BentoMLException("Could not find bentoml module location.")
    bentoml_path = Path(module_location)
    bentoml_project = bentoml_path.parent.parent

    if not (
        bentoml_path / "grpc" / _internal_protocol_version / "service_pb2.py"
    ).exists():
        raise ModuleNotFoundError(
            f"Generated stubs for version {_internal_protocol_version} are missing. Make sure to run '{bentoml_path.parent.as_posix()}/tools/generate-grpc-stubs {_internal_protocol_version}' beforehand to generate gRPC stubs."
        ) from None

    # this is for BentoML developer to create Service containing custom development
    # branches of BentoML library, it is True only when BentoML module is installed
    # in development mode via "pip install --editable ."
    if not (bentoml_project / "pyproject.toml").exists():
        logger.warning(
            "Custom BentoML build is detected. For a Bento to use the same build at serving time, add your custom BentoML build to the pip packages list, e.g. `packages=['git+https://github.com/bentoml/bentoml.git@13dfb36']`"
        )
        return

    logger.debug(
        "BentoML is installed in `editable` mode; building BentoML distribution with the local BentoML code base. The built wheel file will be included in the target bento."
    )

    def exclude_pycache(tarinfo: tarfile.TarInfo):
        if "__pycache__" in tarinfo.name or tarinfo.name.endswith((".pyc", ".pyo")):
            return None
        return tarinfo

    # Freeze the version in pyproject.toml
    with open(bentoml_project / "pyproject.toml", "rb") as f:
        pyproject_toml = tomllib.load(f)

    pyproject_toml["project"]["version"] = BENTOML_VERSION
    if (
        "dynamic" in pyproject_toml["project"]
        and "version" in pyproject_toml["project"]["dynamic"]
    ):
        pyproject_toml["project"]["dynamic"].remove("version")

    pyproject_io = BytesIO()
    tomli_w.dump(pyproject_toml, pyproject_io)

    # Make a tarball of the BentoML source code
    base_name = f"bentoml-{BENTOML_VERSION}"
    sdist_filename = f"{base_name}.tar.gz"
    files_to_include = ["src", "LICENSE", "README.md"]

    if not Path(target_path).exists():
        Path(target_path).mkdir(parents=True, exist_ok=True)

    with tarfile.open(Path(target_path, sdist_filename), "w:gz") as tar:
        for file in files_to_include:
            tar.add(
                bentoml_project / file,
                arcname=f"{base_name}/{file}",
                filter=exclude_pycache,
            )
        tarinfo = tar.gettarinfo(
            bentoml_project / "pyproject.toml", arcname=f"{base_name}/pyproject.toml"
        )
        tarinfo.size = pyproject_io.tell()
        pyproject_io.seek(0)
        tar.addfile(tarinfo, pyproject_io)

    return sdist_filename


def build_git_repo(url: str, ref: str, subdirectory: str | None, dst_path: str) -> str:
    """
    Download the git repo at the given url and ref, and build an sdist from it at dst_path
    """
    import shutil
    import subprocess

    from ...exceptions import BentoMLException
    from ..configuration import get_uv_command

    name = os.path.splitext(os.path.basename(url))[0]

    with fs.open_fs("temp://") as temp_fs:
        dest_dir = temp_fs.getsyspath(name)
        git_command = ["git", "clone", "--filter=blob:none", "--quiet", url, dest_dir]
        env = {**os.environ, "GIT_TERMINAL_PROMPT": "0"}
        try:
            subprocess.check_call(
                git_command,
                env=env,
                stderr=subprocess.PIPE,
                stdout=subprocess.DEVNULL,
            )
            if ref:
                subprocess.check_call(
                    ["git", "fetch", "-q", url, ref],
                    cwd=dest_dir,
                    stderr=subprocess.PIPE,
                    stdout=subprocess.DEVNULL,
                )
                subprocess.check_call(
                    ["git", "checkout", "FETCH_HEAD"],
                    cwd=dest_dir,
                    stderr=subprocess.PIPE,
                    stdout=subprocess.DEVNULL,
                )
        except subprocess.CalledProcessError as e:
            raise BentoMLException(
                f"Failed to clone git repository {url}: {e.stderr}"
            ) from e
        source_dir = os.path.join(dest_dir, subdirectory) if subdirectory else dest_dir
        build_cmd = [*get_uv_command(), "build", "--sdist"]
        subprocess.check_call(build_cmd, cwd=source_dir)
        sdist = next(Path(source_dir).glob("dist/*.tar.gz"))
        logger.info(f"Built sdist {sdist.name}")
        os.makedirs(dst_path, exist_ok=True)
        shutil.move(sdist, dst_path)
        return sdist.name



from __future__ import annotations

import logging
import typing as t
from abc import ABC
from abc import abstractmethod

import attr
from simple_di import Provide
from simple_di import inject

from ...exceptions import StateException
from ..configuration.containers import BentoMLContainer
from ..models.model import Model
from ..tag import validate_tag_str
from ..utils import deprecated
from ..utils import first_not_none
from .runnable import Runnable
from .runner_handle import DummyRunnerHandle
from .runner_handle import RunnerHandle
from .strategy import DefaultStrategy
from .strategy import Strategy

if t.TYPE_CHECKING:
    from ...triton import Runner as TritonRunner
    from .runnable import RunnableMethodConfig

    # only use ParamSpec in type checking, as it's only in 3.10
    P = t.ParamSpec("P")
    ListModel = list[Model]
else:
    P = t.TypeVar("P")
    ListModel = list

T = t.TypeVar("T", bound=Runnable)
R = t.TypeVar("R")


logger = logging.getLogger(__name__)

object_setattr = object.__setattr__


@attr.frozen(slots=False)
class RunnerMethod(t.Generic[T, P, R]):
    runner: Runner | TritonRunner
    name: str
    config: RunnableMethodConfig
    max_batch_size: int
    max_latency_ms: int
    doc: str | None = None

    def run(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self.runner._runner_handle.run_method(self, *args, **kwargs)

    async def async_run(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return await self.runner._runner_handle.async_run_method(self, *args, **kwargs)

    def async_stream(
        self, *args: P.args, **kwargs: P.kwargs
    ) -> t.AsyncGenerator[str, None]:
        return self.runner._runner_handle.async_stream_method(self, *args, **kwargs)


def _to_lower_name(name: str) -> str:
    lname = name.lower()
    if name != lname:
        logger.warning("Converting runner name '%s' to lowercase: '%s'", name, lname)

    return lname


def _validate_name(_: t.Any, attr: attr.Attribute[str], value: str):
    try:
        validate_tag_str(value)
    except ValueError as e:
        # TODO: link to tag validation documentation
        raise ValueError(
            f"Runner name '{value}' is not valid; it must be a valid BentoML Tag name."
        ) from e


@attr.define(slots=False, frozen=True)
class AbstractRunner(ABC):
    name: str = attr.field(converter=_to_lower_name, validator=_validate_name)
    models: list[Model] = attr.field(
        converter=attr.converters.default_if_none(factory=list),
        validator=attr.validators.deep_iterable(
            attr.validators.instance_of(Model),
            iterable_validator=attr.validators.instance_of(ListModel),
        ),
    )
    resource_config: dict[str, t.Any]
    runnable_class: type[Runnable]
    embedded: bool

    @abstractmethod
    def init_local(self, quiet: bool = False) -> None:
        """
        Initialize local runnable instance, for testing and debugging only.

        Args:
            quiet: if True, no logs will be printed
        """

    @abstractmethod
    def init_client(
        self,
        handle_class: type[RunnerHandle] | None = None,
        *args: t.Any,
        **kwargs: t.Any,
    ):
        """
        Initialize client for a remote runner instance. To be used within API server instance.
        """


@deprecated(suggestion="Please upgrade to new style services.")
@attr.define(slots=False, frozen=True, eq=False, init=False)
class Runner(AbstractRunner):
    if t.TYPE_CHECKING:
        # This will be set by __init__. This is for type checking only.
        run: t.Callable[..., t.Any]
        async_run: t.Callable[..., t.Awaitable[t.Any]]

        # the following annotations hacks around the fact that Runner does not
        # have information about signatures at runtime.
        @t.overload
        def __getattr__(
            self, item: t.Literal["__attrs_init__"]
        ) -> t.Callable[..., None]:  # type: ignore
            ...

        @t.overload
        def __getattr__(
            self, item: t.LiteralString
        ) -> RunnerMethod[t.Any, P, t.Any]: ...

        def __getattr__(self, item: str) -> t.Any: ...

    runner_methods: list[RunnerMethod[t.Any, t.Any, t.Any]]
    scheduling_strategy: type[Strategy]
    workers_per_resource: int | float = 1
    runnable_init_params: dict[str, t.Any] = attr.field(
        default=None, converter=attr.converters.default_if_none(factory=dict)
    )
    _runner_handle: RunnerHandle = attr.field(init=False, factory=DummyRunnerHandle)

    def _set_handle(
        self, handle_class: type[RunnerHandle], *args: t.Any, **kwargs: t.Any
    ) -> None:
        if not isinstance(self._runner_handle, DummyRunnerHandle):
            raise StateException("Runner already initialized")

        runner_handle = handle_class(self, *args, **kwargs)
        object_setattr(self, "_runner_handle", runner_handle)

    @inject
    async def runner_handle_is_ready(
        self,
        timeout: int = Provide[BentoMLContainer.api_server_config.runner_probe.timeout],
    ) -> bool:
        """
        Check if given runner handle is ready. This will be used as readiness probe in Kubernetes.
        """
        return await self._runner_handle.is_ready(timeout)

    def __init__(
        self,
        runnable_class: type[Runnable],
        *,
        runnable_init_params: dict[str, t.Any] | None = None,
        name: str | None = None,
        scheduling_strategy: type[Strategy] = DefaultStrategy,
        models: list[Model] | None = None,
        max_batch_size: int | None = None,
        max_latency_ms: int | None = None,
        method_configs: dict[str, dict[str, int]] | None = None,
        embedded: bool = False,
    ) -> None:
        """

        Runner represents a unit of computation that can be executed on a remote Python worker and scales independently
        See https://docs.bentoml.com/en/latest/concepts/runner.html for more details.

        Args:
            runnable_class: Runnable class that can be executed on a remote Python worker.
            runnable_init_params: Parameters to be passed to the runnable class constructor ``__init__``.
            name: Given a name for this runner. If not provided, name will be generated from the runnable class name.
                  Note that all name will be converted to lowercase and validate to be a valid BentoML Tag name.
            scheduling_strategy: A strategy class that implements the scheduling logic for this runner. If not provided,
                                 use the default strategy. Strategy will respect ``Runnable.SUPPORTED_RESOURCES`` as well as
                                 ``Runnable.SUPPORTS_CPU_MULTI_THREADING``.
            models: An optional list composed of ``bentoml.Model`` instances.
            max_batch_size: Max batch size config for dynamic batching. If not provided, use the default value from
                            configuration.
            max_latency_ms: Max latency config for dynamic batching. If not provided, use the default value from
                            configuration.
            method_configs: A dictionary per method config for this given Runner signatures.

        Returns:
            :obj:`bentoml.Runner`: A Runner instance.
        """

        if name is None:
            name = runnable_class.__name__.lower()
            logger.warning(
                "Using lowercased runnable class name '%s' for runner.", name
            )

        runners_config = BentoMLContainer.config.runners.get()
        # If given runner is configured, then use it. Otherwise use the default configuration.
        if name in runners_config:
            config = runners_config[name]
        else:
            config = runners_config

        if models is None:
            models = []

        runner_method_map: dict[str, RunnerMethod[t.Any, t.Any, t.Any]] = {}
        method_configs = {} if method_configs is None else method_configs

        if runnable_class.bentoml_runnable_methods__ is None:
            raise ValueError(
                f"Runnable class '{runnable_class.__name__}' has no methods!"
            )

        for method_name, method in runnable_class.bentoml_runnable_methods__.items():
            if not config["batching"]["enabled"]:
                method.config.batchable = False

            method_max_batch_size = None
            method_max_latency_ms = None
            if method_name in method_configs:
                method_max_batch_size = method_configs[method_name].get(
                    "max_batch_size"
                )
                method_max_latency_ms = method_configs[method_name].get(
                    "max_latency_ms"
                )

            runner_method_map[method_name] = RunnerMethod(
                runner=self,
                name=method_name,
                config=method.config,
                doc=method.func.__doc__,
                max_batch_size=first_not_none(
                    method_max_batch_size,
                    max_batch_size,
                    default=config["batching"]["max_batch_size"],
                ),
                max_latency_ms=first_not_none(
                    method_max_latency_ms,
                    max_latency_ms,
                    default=config["batching"]["max_latency_ms"],
                ),
            )

        self.__attrs_init__(
            name=name,
            models=models,
            runnable_class=runnable_class,
            runnable_init_params=runnable_init_params,
            resource_config=config["resources"],
            workers_per_resource=config.get("workers_per_resource", 1),
            runner_methods=list(runner_method_map.values()),
            scheduling_strategy=scheduling_strategy,
            embedded=embedded,
        )

        # Choose the default method:
        #  1. if there's only one method, it will be set as default
        #  2. if there's a method named "__call__", it will be set as default
        #  3. otherwise, there's no default method
        if len(runner_method_map) == 1:
            default_method = next(iter(runner_method_map.values()))
            logger.debug(
                "Default runner method set to '%s', it can be accessed both via 'runner.run' and 'runner.%s.async_run'.",
                default_method.name,
                default_method.name,
            )
        elif "__call__" in runner_method_map:
            default_method = runner_method_map["__call__"]
            logger.debug(
                "Default runner method set to '__call__', it can be accessed via 'runner.run' or 'runner.async_run'."
            )
        else:
            default_method = None
            logger.debug(
                "No default method found for Runner '%s', all method access needs to be in the form of 'runner.{method}.run'.",
                name,
            )

        # set default run method entrypoint
        if default_method is not None:
            object.__setattr__(self, "run", default_method.run)
            object.__setattr__(self, "async_run", default_method.async_run)

        # set all run method entrypoint
        for runner_method in self.runner_methods:
            object.__setattr__(self, runner_method.name, runner_method)

    def init_local(self, quiet: bool = False) -> None:
        if not quiet:
            logger.warning(
                "'Runner.init_local' is for debugging and testing only. Make sure to remove it before deploying to production."
            )

        from .runner_handle.local import LocalRunnerRef

        try:
            self._set_handle(LocalRunnerRef)
        except Exception as e:
            import traceback

            logger.error(
                "An exception occurred while instantiating runner '%s', see details below:",
                self.name,
            )
            logger.error(traceback.format_exc())

            raise e

    def init_client(
        self,
        handle_class: type[RunnerHandle] | None = None,
        *args: t.Any,
        **kwargs: t.Any,
    ):
        if handle_class is None:
            from .runner_handle.remote import RemoteRunnerClient

            self._set_handle(RemoteRunnerClient)
        else:
            self._set_handle(handle_class, *args, **kwargs)

    def destroy(self):
        """
        Destroy the runner. This is called when the runner is no longer needed.
        Currently used under ``on_shutdown`` event of the BentoML server.
        """
        object_setattr(self, "_runner_handle", DummyRunnerHandle())

    @property
    def scheduled_worker_count(self) -> int:
        return self.scheduling_strategy.get_worker_count(
            self.runnable_class,
            self.resource_config,
            self.workers_per_resource,
        )

    @property
    def scheduled_worker_env_map(self) -> dict[int, dict[str, t.Any]]:
        return {
            worker_id: self.scheduling_strategy.get_worker_env(
                self.runnable_class,
                self.resource_config,
                self.workers_per_resource,
                worker_id,
            )
            for worker_id in range(self.scheduled_worker_count)
        }



from __future__ import annotations

import json
import logging
import os
import typing as t
from datetime import datetime
from datetime import timezone
from typing import TYPE_CHECKING

import attr
import fs
import fs.errors
import fs.mirror
import fs.osfs
import fs.walk
import yaml
from cattr.gen import make_dict_structure_fn
from cattr.gen import make_dict_unstructure_fn
from cattr.gen import override
from fs.copy import copy_file
from fs.tempfs import TempFS
from simple_di import Provide
from simple_di import inject

from ...exceptions import BentoMLException
from ...exceptions import InvalidArgument
from ...exceptions import NotFound
from ..configuration import BENTOML_VERSION
from ..configuration.containers import BentoMLContainer
from ..models import ModelStore
from ..models import copy_model
from ..runner import Runner
from ..store import Store
from ..store import StoreItem
from ..tag import Tag
from ..tag import to_snake_case
from ..types import PathType
from ..utils import normalize_labels_value
from ..utils.cattr import bentoml_cattr
from ..utils.filesystem import copy_file_to_fs_folder
from ..utils.uri import encode_path_for_uri
from .build_config import BentoBuildConfig
from .build_config import BentoEnvSchema
from .build_config import BentoPathSpec
from .build_config import CondaOptions
from .build_config import DockerOptions
from .build_config import PythonOptions

if TYPE_CHECKING:
    from fs.base import FS

    from _bentoml_sdk import Service as NewService
    from _bentoml_sdk.models import Model
    from _bentoml_sdk.service import ServiceConfig
    from _bentoml_sdk.service.dependency import Dependency

    from ..cloud.schemas.modelschemas import BentoManifestSchema
    from ..models import Model as StoredModel
    from ..service import Service
    from ..service.inference_api import InferenceAPI
else:
    ServiceConfig = t.Dict[str, t.Any]

logger = logging.getLogger(__name__)

BENTO_YAML_FILENAME = "bento.yaml"
BENTO_PROJECT_DIR_NAME = "src"
BENTO_README_FILENAME = "README.md"
DEFAULT_BENTO_BUILD_FILES = ("bentofile.yaml", "pyproject.toml")

API_INFO_MD = "| POST [`/{api}`](#{link}) | {input} | {output} |"

INFERENCE_TABLE_MD = """\
| InferenceAPI | Input | Output |
| ------------ | ----- | ------ |
{content}
"""


def create_inference_api_table(svc: Service) -> str:
    from ..service.openapi import APP_TAG

    contents = [
        API_INFO_MD.format(
            api=api.name,
            link=f"operations-{APP_TAG.name.replace(' ', '_')}-{svc.name}__{api.name}",  # follows operationId from OpenAPI
            input=api.input.__class__.__name__,
            output=api.output.__class__.__name__,
        )
        for api in svc.apis.values()
    ]
    return INFERENCE_TABLE_MD.format(content="\n".join(contents))


def get_default_svc_readme(
    svc: Service | NewService[t.Any], svc_version: str | None = None
) -> str:
    from ..service import Service

    if svc.bento:
        bentoml_version = svc.bento.info.bentoml_version
    else:
        bentoml_version = BENTOML_VERSION

    is_legacy = isinstance(svc, Service)

    if not svc_version:
        if svc.bento and svc.bento.tag.version:
            svc_version = svc.bento.tag.version
        else:
            svc_version = "dev"

    doc = f"""\
# {svc.name}:{svc_version}

[![pypi_status](https://img.shields.io/badge/BentoML-{bentoml_version}-informational)](https://pypi.org/project/BentoML)
[![documentation_status](https://readthedocs.org/projects/bentoml/badge/?version=latest)](https://docs.bentoml.com/)
[![join_slack](https://badgen.net/badge/Join/BentoML%20Slack/cyan?icon=slack)](https://l.bentoml.com/join-slack-swagger)
[![BentoML GitHub Repo](https://img.shields.io/github/stars/bentoml/bentoml?style=social)](https://github.com/bentoml/BentoML)
[![Twitter Follow](https://img.shields.io/twitter/follow/bentomlai?label=Follow%20BentoML&style=social)](https://twitter.com/bentomlai)

This is a Machine Learning Service created with BentoML."""

    if is_legacy and svc.apis:
        doc += f"\n{create_inference_api_table(svc)}\n\n"

    doc += """

## Help

* [📖 Documentation](https://docs.bentoml.com/en/latest/): Learn how to use BentoML.
* [💬 Community](https://l.bentoml.com/join-slack-swagger): Join the BentoML Slack community.
* [🐛 GitHub Issues](https://github.com/bentoml/BentoML/issues): Report bugs and feature requests.
* Tip: you can also [customize this README](https://docs.bentoml.com/en/latest/reference/bentoml/bento-build-options.html#description).
"""
    # TODO: add links to documentation that may help with API client development
    return doc


@attr.define(repr=False, auto_attribs=False)
class Bento(StoreItem):
    _tag: Tag = attr.field()
    __fs: FS = attr.field()

    _info: BaseBentoInfo

    _model_store: ModelStore | None = None
    _doc: str | None = None

    @staticmethod
    def _export_ext() -> str:
        return "bento"

    @__fs.validator  # type:ignore # attrs validators not supported by pyright
    def check_fs(self, _attr: t.Any, new_fs: FS):
        try:
            models = new_fs.opendir("models")
        except fs.errors.ResourceNotFound:
            return
        else:
            self._model_store = ModelStore(models)

    def __init__(self, tag: Tag, bento_fs: "FS", info: "BentoInfo"):
        self._tag = tag
        self.__fs = bento_fs
        self.check_fs(None, bento_fs)
        self._info = info

    @property
    def tag(self) -> Tag:
        return self._tag

    @property
    def _fs(self) -> FS:
        return self.__fs

    @property
    def info(self) -> BaseBentoInfo:
        return self._info

    @property
    def entry_service(self) -> str:
        return self.info.entry_service

    def get_manifest(self, dev: bool = False) -> BentoManifestSchema:
        from ..cloud.schemas.modelschemas import BentoManifestSchema
        from ..cloud.schemas.modelschemas import BentoRunnerResourceSchema
        from ..cloud.schemas.modelschemas import BentoRunnerSchema

        info = self.info
        models = [str(m.tag) for m in info.all_models]
        if isinstance(info, BentoInfo):
            runners = [
                BentoRunnerSchema(
                    name=r.name,
                    runnable_type=r.runnable_type,
                    models=r.models,
                    resource_config=(
                        BentoRunnerResourceSchema(
                            cpu=r.resource_config.get("cpu"),
                            nvidia_gpu=r.resource_config.get("nvidia.com/gpu"),
                            custom_resources=r.resource_config.get("custom_resources"),
                        )
                        if r.resource_config
                        else None
                    ),
                )
                for r in info.runners
            ]
            image: ImageInfo | None = None
        else:
            image = t.cast(ImageInfo, info.image)
            runners = []
        return BentoManifestSchema(
            name=info.name,
            entry_service=info.entry_service,
            service=info.service,
            bentoml_version=info.bentoml_version,
            apis={},
            models=models,
            runners=runners,
            size_bytes=self.total_size(),
            services=info.services,
            envs=info.envs,
            schema=info.schema,
            version=info.version,
            dev=dev,
            image=image,
            spec=info.spec,
        )

    @classmethod
    @inject
    def create(
        cls,
        build_config: BentoBuildConfig,
        version: t.Optional[str] = None,
        build_ctx: t.Optional[str] = None,
        platform: t.Optional[str] = None,
        bare: bool = False,
        reload: bool = False,
        enabled_features: list[str] = Provide[BentoMLContainer.enabled_features],
    ) -> Bento:
        from _bentoml_sdk.images import Image
        from _bentoml_sdk.images import get_image_from_build_config
        from _bentoml_sdk.models import BentoModel

        from ..service import Service
        from ..service.loader import load

        build_ctx = (
            os.getcwd()
            if build_ctx is None
            else os.path.realpath(os.path.expanduser(build_ctx))
        )
        if not os.path.isdir(build_ctx):
            raise InvalidArgument(
                f"Bento build context {build_ctx} does not exist or is not a directory."
            )

        BentoMLContainer.model_aliases.set(build_config.model_aliases)
        # This also verifies that svc can be imported correctly
        svc = load(build_config.service or ".", working_dir=build_ctx, reload=reload)
        if not build_config.service:
            object.__setattr__(build_config, "service", get_service_import_str(svc))
        is_legacy = isinstance(svc, Service)
        # Apply default build options
        image: Image | None = None
        disable_image = "no_image" in enabled_features or is_legacy

        if isinstance(svc, Service):
            # for < 1.2
            bento_name = (
                build_config.name if build_config.name is not None else svc.name
            )
        else:
            # for >= 1.2
            svc.inject_config()
            bento_name = (
                build_config.name
                if build_config.name is not None
                else to_snake_case(svc.name)
            )
            build_config.envs.extend(svc.envs)
            build_config.labels.update(svc.labels)
            if svc.image is not None:
                image = svc.image
        if image is None and not disable_image:
            image = get_image_from_build_config(build_config, build_ctx)
        build_config = build_config.with_defaults()
        tag = Tag(bento_name, version)
        if version is None:
            tag = tag.make_new_version()

        logger.debug(
            'Building BentoML service "%s" from build context "%s".', tag, build_ctx
        )
        bento_fs = TempFS(
            identifier=f"bentoml_bento_{bento_name}",
            temp_dir=BentoMLContainer.tmp_bento_store_dir.get(),
        )
        models: list[BentoModelInfo] = []

        def append_model(model: BentoModelInfo) -> None:
            if model not in models:
                models.append(model)

        if build_config.models:
            for model_spec in build_config.models:
                model = BentoModel(model_spec.tag)
                append_model(model.to_info(model_spec.alias))
        elif is_legacy:
            # XXX: legacy way to get models from service
            # Add all models required by the service
            for model in svc.models:
                append_model(BentoModel(model.tag).to_info())
            # Add all models required by service runners
            for runner in svc.runners:
                for model in runner.models:
                    append_model(BentoModel(model.tag).to_info())

        if not bare:
            ctx_fs = fs.open_fs(encode_path_for_uri(build_ctx))

            # create ignore specs
            specs = BentoPathSpec(build_config.include, build_config.exclude, build_ctx)

            # Copy all files base on include and exclude, into `src` directory
            relpaths = [s for s in build_config.include if s.startswith("../")]
            if len(relpaths) != 0:
                raise InvalidArgument(
                    "Paths outside of the build context directory cannot be included; use a symlink or copy those files into the working directory manually."
                )
            bento_fs.makedir(BENTO_PROJECT_DIR_NAME)
            target_fs = bento_fs.opendir(BENTO_PROJECT_DIR_NAME)
            with target_fs.open("bentofile.yaml", "w") as bentofile_yaml:
                build_config.to_yaml(bentofile_yaml)

            for dir_path, _, files in ctx_fs.walk():
                for f in files:
                    path = fs.path.combine(dir_path, f.name).lstrip("/")
                    if specs.includes(path):
                        if ctx_fs.getsize(path) > 10 * 1024 * 1024:
                            logger.warn("File size is larger than 10MiB: %s", path)
                        target_fs.makedirs(dir_path, recreate=True)
                        copy_file(ctx_fs, path, target_fs, path)
            if image is None:
                # NOTE: we need to generate both Python and Conda
                # first to make sure we can generate the Dockerfile correctly.
                build_config.python.write_to_bento(
                    bento_fs, build_ctx, platform_=platform
                )
                build_config.conda.write_to_bento(bento_fs, build_ctx)
                build_config.docker.write_to_bento(
                    bento_fs, build_ctx, build_config.conda
                )

            # Create `readme.md` file
            if (
                build_config.description is not None
                and build_config.description.startswith("file:")
            ):
                file_name = build_config.description[5:].strip()
                if not ctx_fs.exists(file_name):
                    raise InvalidArgument(f"File {file_name} does not exist.")
                copy_file_to_fs_folder(
                    ctx_fs.getsyspath(file_name),
                    bento_fs,
                    dst_filename=BENTO_README_FILENAME,
                )
            elif build_config.description is None and ctx_fs.exists(
                BENTO_README_FILENAME
            ):
                copy_file_to_fs_folder(
                    ctx_fs.getsyspath(BENTO_README_FILENAME),
                    bento_fs,
                    dst_filename=BENTO_README_FILENAME,
                )
            else:
                with bento_fs.open(BENTO_README_FILENAME, "w") as f:
                    if build_config.description is None:
                        f.write(get_default_svc_readme(svc, version))
                    else:
                        f.write(build_config.description)

            # Create 'apis/openapi.yaml' file
            bento_fs.makedir("apis")
            with bento_fs.open(fs.path.combine("apis", "openapi.yaml"), "w") as f:
                yaml.dump(svc.openapi_spec, f)
            if not is_legacy:
                with bento_fs.open(fs.path.combine("apis", "schema.json"), "w") as f:
                    json.dump(svc.schema(), f, indent=2)

        if image is None:
            bento_info = BentoInfo(
                tag=tag,
                service=svc,  # type: ignore # attrs converters do not typecheck
                entry_service=svc.name,
                labels=build_config.labels,
                models=models,
                runners=(
                    [BentoRunnerInfo.from_runner(r) for r in svc.runners]  # type: ignore # attrs converters do not typecheck
                    if is_legacy
                    else []
                ),
                apis=(
                    [BentoApiInfo.from_inference_api(api) for api in svc.apis.values()]
                    if is_legacy
                    else []
                ),
                services=(
                    [
                        BentoServiceInfo.from_service(s)
                        for s in svc.all_services().values()
                    ]
                    if not is_legacy
                    else []
                ),
                docker=build_config.docker,
                python=build_config.python,
                conda=build_config.conda,
                envs=build_config.envs,
                schema=svc.schema() if not is_legacy else {},
            )
        else:
            bento_info = BentoInfoV2(
                tag=tag,
                service=svc,  # type: ignore # attrs converters do not typecheck
                entry_service=svc.name,
                labels=build_config.labels,
                models=models,
                services=(
                    [
                        BentoServiceInfo.from_service(s)
                        for s in svc.all_services().values()
                    ]
                    if not is_legacy
                    else []
                ),
                envs=build_config.envs,
                schema=svc.schema() if not is_legacy else {},
                image=image.freeze(platform),
            )
            bento_info.image.write_to_bento(bento_fs, build_config.envs)

        res = Bento(tag, bento_fs, bento_info)
        if bare:
            return res
        # Create bento.yaml
        res.flush_info()
        try:
            res.validate()
        except BentoMLException as e:
            raise BentoMLException(f"Failed to create {res!s}: {e}") from None

        return res

    @classmethod
    def from_fs(cls, item_fs: FS) -> Bento:
        try:
            with item_fs.open(BENTO_YAML_FILENAME, "r", encoding="utf-8") as bento_yaml:
                info = BaseBentoInfo.from_yaml_file(bento_yaml)
        except fs.errors.ResourceNotFound:
            raise BentoMLException(
                f"Failed to load bento because it does not contain a '{BENTO_YAML_FILENAME}'"
            )

        res = cls(info.tag, item_fs, info)
        try:
            res.validate()
        except BentoMLException as e:
            raise BentoMLException(f"Failed to load bento: {e}") from None

        return res

    @classmethod
    def from_path(cls, item_path: str) -> Bento:
        item_path = os.path.expanduser(item_path)
        item_fs = fs.open_fs(encode_path_for_uri(item_path))
        return cls.from_fs(item_fs)

    def export(
        self,
        path: str,
        output_format: t.Optional[str] = None,
        *,
        protocol: t.Optional[str] = None,
        user: t.Optional[str] = None,
        passwd: t.Optional[str] = None,
        params: t.Optional[t.Dict[str, str]] = None,
        subpath: t.Optional[str] = None,
    ) -> str:
        from _bentoml_sdk.models import BentoModel

        # copy the models to fs
        models_dir = self._fs.makedir("models", recreate=True)
        for model in self.info.all_models:
            if model.registry != "bentoml":
                continue
            BentoModel(model.tag).resolve(models_dir)
        try:
            return super().export(
                path,
                output_format,
                protocol=protocol,
                user=user,
                passwd=passwd,
                params=params,
                subpath=subpath,
            )
        finally:
            self._fs.removetree("models")

    @inject
    def total_size(
        self, model_store: ModelStore = Provide[BentoMLContainer.model_store]
    ) -> int:
        total_size = self.file_size
        local_model_store = self._model_store
        for model in self.info.all_models:
            if model.registry == "huggingface":
                continue
            if local_model_store is not None:
                try:
                    local_model_store.get(model.tag)
                    continue
                except NotFound:
                    pass
            try:
                global_model = model_store.get(model.tag)
            except NotFound:
                logger.warning(
                    "Bento: %s: Missing model %s",
                    self.tag,
                    model.tag,
                )
                continue
            total_size += global_model.file_size
        return total_size

    def flush_info(self):
        with self._fs.open(BENTO_YAML_FILENAME, "w") as bento_yaml:
            self.info.dump(bento_yaml)

    @property
    def doc(self) -> str:
        if self._doc is not None:
            return self._doc
        if not self._fs.isfile(BENTO_README_FILENAME):
            return ""
        with self._fs.open(BENTO_README_FILENAME, "r") as readme_md:
            self._doc = str(readme_md.read())
            return self._doc

    @property
    def creation_time(self) -> datetime:
        return self.info.creation_time

    @inject
    def save(
        self,
        bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ) -> "Bento":
        try:
            self.validate()
        except BentoMLException as e:
            raise BentoMLException(f"Failed to save {self!s}: {e}") from None

        with bento_store.register(self.tag) as bento_path:
            out_fs = fs.open_fs(
                encode_path_for_uri(bento_path), create=True, writeable=True
            )
            if self._model_store is not None:
                # Move models to the global model store, if any
                for model in self._model_store.list():
                    copy_model(
                        model.tag,
                        src_model_store=self._model_store,
                        target_model_store=model_store,
                    )
                self._model_store = None
            fs.mirror.mirror(self._fs, out_fs, copy_if_newer=False)
            try:
                out_fs.removetree("models")
            except fs.errors.ResourceNotFound:
                pass
            self._fs.close()
            self.__fs = out_fs

        return self

    def validate(self):
        if not self._fs.isfile(BENTO_YAML_FILENAME):
            raise BentoMLException(
                f"{self!s} does not contain a {BENTO_YAML_FILENAME}."
            )

    def __str__(self):
        return f'Bento(tag="{self.tag}")'


class BentoStore(Store[Bento]):
    def __init__(self, base_path: t.Union[PathType, "FS"]):
        super().__init__(base_path, Bento)


@attr.frozen
class BentoRunnerInfo:
    name: str
    runnable_type: str
    embedded: bool = attr.field(default=False)
    models: t.List[str] = attr.field(factory=list)
    resource_config: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)

    @classmethod
    def from_runner(cls, r: Runner) -> BentoRunnerInfo:
        return cls(
            name=r.name,
            runnable_type=r.runnable_class.__name__,
            embedded=r.embedded,
            models=[str(model.tag) for model in r.models],
            resource_config=r.resource_config,
        )


@attr.frozen
class BentoApiInfo:
    name: str
    input_type: str
    output_type: str

    @classmethod
    def from_inference_api(cls, api: InferenceAPI[t.Any]) -> BentoApiInfo:
        return cls(
            name=api.name,
            input_type=api.input.__class__.__name__,
            output_type=api.output.__class__.__name__,
        )


@attr.frozen
class BentoModelInfo:
    tag: Tag = attr.field(converter=Tag.from_taglike)
    module: t.Optional[str] = attr.field(default=None, eq=False)
    creation_time: datetime = attr.field(
        factory=lambda: datetime.now(timezone.utc), eq=False
    )
    alias: t.Optional[str] = attr.field(default=None, eq=False)
    registry: str = attr.field(default="bentoml", eq=False)
    metadata: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None, eq=False)

    @classmethod
    def from_bento_model(
        cls, bento_model: StoredModel, alias: str | None = None
    ) -> BentoModelInfo:
        return cls(
            tag=bento_model.tag,
            module=bento_model.info.module,
            creation_time=bento_model.info.creation_time,
            alias=alias,
        )

    def to_model(self) -> Model[t.Any]:
        from bentoml.models import BentoModel
        from bentoml.models import HuggingFaceModel

        if self.registry == "bentoml":
            return BentoModel.from_info(self)
        elif self.registry == "huggingface":
            return HuggingFaceModel.from_info(self)
        else:
            raise ValueError(f"Unknown model registry: {self.registry}")


@attr.frozen
class BentoDependencyInfo:
    service: t.Optional[str] = None
    deployment: t.Optional[str] = None
    cluster: t.Optional[str] = None
    url: t.Optional[str] = None

    @classmethod
    def from_dependency(cls, d: Dependency[t.Any]) -> BentoDependencyInfo:
        return cls(
            service=d.on.name if d.on is not None else None,
            deployment=d.deployment,
            cluster=d.cluster,
            url=d.url,
        )


@attr.frozen
class BentoServiceInfo:
    name: str
    service: str
    models: t.List[BentoModelInfo] = attr.field(factory=list, eq=False)
    dependencies: t.List[BentoDependencyInfo] = attr.field(factory=list, eq=False)
    config: ServiceConfig = attr.field(factory=dict, eq=False)

    @classmethod
    def from_service(cls, svc: NewService[t.Any]) -> BentoServiceInfo:
        return cls(
            name=svc.name,
            service="",
            models=[m.to_info() for m in svc.models],
            dependencies=[
                BentoDependencyInfo.from_dependency(d)
                for d in svc.dependencies.values()
            ],
            config=svc.config,
        )


def get_service_import_str(svc: Service | NewService[t.Any] | str) -> str:
    from ..service import Service

    if isinstance(svc, Service):
        return svc.get_service_import_origin()[0]
    if isinstance(svc, str):
        return svc
    return svc.import_string


@attr.frozen(repr=False)
class BaseBentoInfo:
    # for backward compatibility in case new fields are added to BentoInfo.
    __forbid_extra_keys__ = False
    # omit field in yaml file if it is not provided by the user.
    __omit_if_default__ = True

    tag: Tag
    service: str = attr.field(converter=get_service_import_str)
    name: str = attr.field(init=False)
    version: str = attr.field(init=False)
    # using factory explicitly instead of default because omit_if_default is enabled for BentoInfo
    bentoml_version: str = attr.field(factory=lambda: BENTOML_VERSION)
    creation_time: datetime = attr.field(factory=lambda: datetime.now(timezone.utc))

    labels: t.Dict[str, t.Any] = attr.field(
        factory=dict, converter=normalize_labels_value
    )
    models: t.List[BentoModelInfo] = attr.field(factory=list)
    # for BentoML 1.2+ SDK
    entry_service: str = attr.field(factory=str)
    services: t.List[BentoServiceInfo] = attr.field(factory=list)
    envs: t.List[BentoEnvSchema] = attr.field(factory=list)
    schema: t.Dict[str, t.Any] = attr.field(factory=dict)

    @property
    def all_models(self) -> t.List[BentoModelInfo]:
        model_map = {model.tag: model for model in self.models}
        for service in self.services:
            for model in service.models:
                model_map[model.tag] = model
        return list(model_map.values())

    def __attrs_post_init__(self):
        # Direct set is not available when frozen=True
        object.__setattr__(self, "name", self.tag.name)
        object.__setattr__(self, "version", self.tag.version)

        try:
            self.validate()
        except BentoMLException as e:
            raise BentoMLException(f"Failed to initialize {self!s}: {e}") from None

    def to_dict(self) -> t.Dict[str, t.Any]:
        return bentoml_cattr.unstructure(self)

    def dump(self, stream: t.IO[t.Any]) -> None:
        # _models is an alias for models, replace it with models
        return yaml.safe_dump(self.to_dict(), stream, sort_keys=False)

    @classmethod
    def from_yaml_file(cls, stream: t.IO[t.Any]) -> BaseBentoInfo:
        try:
            yaml_content = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            logger.error("Error while parsing YAML file: %s", exc)
            raise

        assert yaml_content is not None

        yaml_content["tag"] = Tag(yaml_content["name"], yaml_content["version"])
        del yaml_content["name"]
        del yaml_content["version"]

        # For backwards compatibility for bentos created prior to version 1.0.0rc1
        if "runners" in yaml_content:
            runners = yaml_content["runners"]
            for r in runners:
                if "runner_type" in r:  # BentoRunnerInfo prior to 1.0.0rc1 release
                    r["runnable_type"] = r["runner_type"]
                    del r["runner_type"]
                    if "model_runner_module" in r:
                        del r["model_runner_module"]

        if "models" in yaml_content:
            # For backwards compatibility for bentos created prior to version 1.0.0a7
            models = yaml_content["models"]
            if models and len(models) > 0 and isinstance(models[0], str):
                yaml_content["models"] = list(
                    map(
                        lambda model_tag: {
                            "tag": model_tag,
                            "module": "unknown",
                            "creation_time": datetime.fromordinal(1),
                        },
                        models,
                    )
                )
        if yaml_content.pop("spec", 1) == 2:
            klass = BentoInfoV2
        else:
            klass = BentoInfo
        try:
            return bentoml_cattr.structure(yaml_content, klass)
        except KeyError as e:
            raise BentoMLException(f"Missing field {e} in {BENTO_YAML_FILENAME}")

    def validate(self):
        # Validate bento.yml file schema, content, bentoml version, etc
        ...


@attr.frozen(repr=False)
class BentoInfo(BaseBentoInfo):
    spec: int = attr.field(default=1, init=False)
    runners: t.List[BentoRunnerInfo] = attr.field(factory=list)
    apis: t.List[BentoApiInfo] = attr.field(factory=list)
    docker: DockerOptions = attr.field(factory=lambda: DockerOptions().with_defaults())
    python: PythonOptions = attr.field(factory=lambda: PythonOptions().with_defaults())
    conda: CondaOptions = attr.field(factory=lambda: CondaOptions().with_defaults())


@attr.frozen(repr=False)
class ImageInfo:
    base_image: str = ""
    python_version: str = ""
    commands: t.List[str] = attr.field(factory=list)
    python_requirements: str = ""
    post_commands: t.List[str] = attr.field(factory=list)
    scripts: t.Dict[str, str] = attr.field(factory=dict)

    def write_to_bento(self, bento_fs: FS, envs: list[BentoEnvSchema]) -> None:
        from importlib import resources

        from _bentoml_impl.docker import generate_dockerfile

        # Prepare env/python files
        py_folder = fs.path.join("env", "python")
        bento_fs.makedirs(py_folder, recreate=True)
        reqs_txt = fs.path.join(py_folder, "requirements.txt")
        bento_fs.writetext(reqs_txt, self.python_requirements)
        # Prepare env/docker files
        docker_folder = fs.path.join("env", "docker")
        bento_fs.makedirs(docker_folder, recreate=True)
        dockerfile_path = fs.path.join(docker_folder, "Dockerfile")
        bento_fs.writetext(
            dockerfile_path,
            generate_dockerfile(self, bento_fs, enable_buildkit=False, envs=envs),
        )
        for script_name, target_path in self.scripts.items():
            copy_file_to_fs_folder(script_name, bento_fs, dst_filename=target_path)

        with resources.path(
            "bentoml._internal.container.frontend.dockerfile", "entrypoint.sh"
        ) as entrypoint_path:
            copy_file_to_fs_folder(str(entrypoint_path), bento_fs, docker_folder)


@attr.frozen(repr=False)
class BentoInfoV2(BaseBentoInfo):
    spec: int = attr.field(default=2, init=False)
    image: ImageInfo = attr.field(factory=ImageInfo)


bentoml_cattr.register_unstructure_hook(
    BentoDependencyInfo,
    make_dict_unstructure_fn(
        BentoDependencyInfo,
        bentoml_cattr,
        service=override(omit_if_default=True),
        deployment=override(omit_if_default=True),
        cluster=override(omit_if_default=True),
        url=override(omit_if_default=True),
    ),
)


def _convert_bento_dependency_info(
    data: str | dict[str, t.Any], typ: type[BentoDependencyInfo]
) -> BentoDependencyInfo:
    if isinstance(data, str):
        return BentoDependencyInfo(service=data)
    return BentoDependencyInfo(**data)


bentoml_cattr.register_structure_hook(
    BentoDependencyInfo, _convert_bento_dependency_info
)

bentoml_cattr.register_unstructure_hook(
    BentoModelInfo,
    make_dict_unstructure_fn(
        BentoModelInfo,
        bentoml_cattr,
        alias=override(omit_if_default=True),
        registry=override(omit_if_default=True),
        metadata=override(omit_if_default=True),
    ),
)
bentoml_cattr.register_structure_hook_factory(
    lambda cls: issubclass(cls, BaseBentoInfo),
    lambda cls: make_dict_structure_fn(
        cls,
        bentoml_cattr,
        name=override(omit=True),
        version=override(omit=True),
    ),
)
bentoml_cattr.register_unstructure_hook_factory(
    lambda cls: issubclass(cls, BaseBentoInfo),
    # Ignore tag, tag is saved via the name and version field
    lambda cls: make_dict_unstructure_fn(cls, bentoml_cattr, tag=override(omit=True)),
)



from __future__ import annotations

import json
import logging
import os
import typing as t
from http import HTTPStatus

import click
import rich
import rich.style
import yaml
from rich.console import Console
from rich.syntax import Syntax
from rich.table import Table
from simple_di import Provide
from simple_di import inject

import bentoml.deployment
from bentoml._internal.cloud.base import Spinner
from bentoml._internal.cloud.client import RestApiClient
from bentoml._internal.cloud.deployment import Deployment
from bentoml._internal.cloud.deployment import DeploymentConfigParameters
from bentoml._internal.cloud.schemas.modelschemas import DeploymentStatus
from bentoml._internal.cloud.schemas.modelschemas import DeploymentStrategy
from bentoml._internal.configuration.containers import BentoMLContainer
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import CLIException
from bentoml_cli.utils import BentoMLCommandGroup

logger = logging.getLogger("bentoml.cli.deployment")

if t.TYPE_CHECKING:
    from bentoml._internal.cloud import BentoCloudClient

    TupleStrAny = tuple[str, ...]
else:
    TupleStrAny = tuple


def raise_deployment_config_error(err: BentoMLException, action: str) -> t.NoReturn:
    if err.error_code == HTTPStatus.UNAUTHORIZED:
        raise BentoMLException(
            f"{err}\n* BentoCloud API token is required for authorization. Run `bentoml cloud login` command to login"
        ) from None
    raise BentoMLException(
        f"Failed to {action} deployment due to invalid configuration: {err}"
    ) from None


def convert_env_to_dict(env: tuple[str] | None) -> list[dict[str, str]] | None:
    if env is None:
        return None
    collected_envs: list[dict[str, str]] = []
    if env:
        for item in env:
            if "=" in item:
                name, value = item.split("=", 1)
            else:
                name = item
                if name not in os.environ:
                    raise CLIException(f"Environment variable {name} not found")
                value = os.environ[name]
            collected_envs.append({"name": name, "value": value})
    return collected_envs


@click.command(name="deploy")
@click.argument(
    "bento",
    type=click.STRING,
    required=False,
)
@click.option(
    "-n",
    "--name",
    type=click.STRING,
    help="Deployment name",
)
@click.option(
    "--cluster",
    type=click.STRING,
    help="Name of the cluster",
)
@click.option(
    "--access-authorization",
    type=click.BOOL,
    help="Enable access authorization",
)
@click.option(
    "--scaling-min",
    type=click.INT,
    help="Minimum scaling value",
)
@click.option(
    "--scaling-max",
    type=click.INT,
    help="Maximum scaling value",
)
@click.option(
    "--instance-type",
    type=click.STRING,
    help="Type of instance",
)
@click.option(
    "--strategy",
    type=click.Choice(
        [deployment_strategy.value for deployment_strategy in DeploymentStrategy]
    ),
    help="Deployment strategy",
)
@click.option(
    "--env",
    type=click.STRING,
    help="List of environment variables pass by --env key[=value] --env ...",
    multiple=True,
)
@click.option(
    "--secret",
    type=click.STRING,
    help="List of secret names pass by --secret name1, --secret name2, ...",
    multiple=True,
)
@click.option(
    "-f",
    "--config-file",
    type=click.File(),
    help="Configuration file path",
    default=None,
)
@click.option(
    "--config-dict",
    type=click.STRING,
    help="Configuration json string",
    default=None,
)
@click.option(
    "--wait/--no-wait",
    type=click.BOOL,
    is_flag=True,
    help="Do not wait for deployment to be ready",
    default=True,
)
@click.option(
    "--timeout",
    type=click.INT,
    default=3600,
    help="Timeout for deployment to be ready in seconds",
)
def deploy_command(
    bento: str | None,
    name: str | None,
    cluster: str | None,
    access_authorization: bool | None,
    scaling_min: int | None,
    scaling_max: int | None,
    instance_type: str | None,
    strategy: str | None,
    env: tuple[str] | None,
    secret: tuple[str] | None,
    config_file: str | t.TextIO | None,
    config_dict: str | None,
    wait: bool,
    timeout: int,
) -> None:
    """Create a deployment on BentoCloud.

    \b
    Create a deployment using parameters, or using config yaml file.
    """
    create_deployment(
        bento=bento,
        name=name,
        cluster=cluster,
        access_authorization=access_authorization,
        scaling_min=scaling_min,
        scaling_max=scaling_max,
        instance_type=instance_type,
        strategy=strategy,
        env=env,
        secret=secret,
        config_file=config_file,
        config_dict=config_dict,
        wait=wait,
        timeout=timeout,
    )


output_option = click.option(
    "-o",
    "--output",
    type=click.Choice(["yaml", "json"]),
    default="yaml",
    help="Display the output of this command.",
)


def shared_decorator(
    f: t.Callable[..., t.Any] | None = None,
) -> t.Callable[..., t.Any]:
    def decorate(f: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        options = [
            click.option(
                "--cluster",
                type=click.STRING,
                default=None,
                help="Name of the cluster.",
            ),
        ]
        for opt in reversed(options):
            f = opt(f)
        return f

    if f:
        return decorate(f)
    else:
        return decorate


@click.command(name="code")
@click.argument(
    "bento_dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True),
    default=".",
)
@click.option(
    "--attach", help="Attach to the given deployment instead of creating a new one."
)
@click.option(
    "--env",
    type=click.STRING,
    help="List of environment variables pass by --env key[=value] --env ...",
    multiple=True,
)
@click.option(
    "--secret",
    type=click.STRING,
    help="List of secret names pass by --secret name1, --secret name2, ...",
    multiple=True,
)
@shared_decorator
@inject
def develop_command(
    bento_dir: str,
    cluster: str | None,
    attach: str | None,
    env: tuple[str] | None,
    secret: tuple[str] | None,
    _rest_client: RestApiClient = Provide[BentoMLContainer.rest_api_client],
):
    """Create or attach to a codespace.

    Create a new codespace:

        $ bentoml code

    Attach to an existing codespace:

        $ bentoml code --attach <codespace-name>
    """
    import questionary

    if attach and (env or secret):
        raise CLIException("Cannot specify both --attach and --env or --secret")

    console = Console(highlight=False)
    if attach:
        deployment = bentoml.deployment.get(attach)
    else:
        with console.status("Fetching codespaces..."):
            current_user = _rest_client.v1.get_current_user()
            if current_user is None:
                raise CLIException("current user is not found")
            deployments = [
                d
                for d in bentoml.deployment.list(cluster=cluster, dev=True)
                if d.is_dev
                and d.created_by == current_user.name
                and d.get_status(False).status
                in [
                    DeploymentStatus.Deploying.value,
                    DeploymentStatus.Running.value,
                    DeploymentStatus.ScaledToZero.value,
                    DeploymentStatus.Failed.value,
                ]
            ]

        chosen = questionary.select(
            message="Select a codespace to attach to or create a new one",
            choices=[{"name": d.name, "value": d} for d in deployments]
            + [{"name": "Create a new codespace", "value": "new"}],
        ).ask()

        if chosen == "new":
            deployment = create_deployment(
                bento=bento_dir,
                cluster=cluster,
                dev=True,
                wait=False,
                env=env,
                secret=secret,
            )
        elif chosen is None:
            return
        else:
            if env or secret:
                rich.print(
                    "[yellow]Warning:[/] --env and --secret are ignored when attaching to an existing codespace"
                )
            deployment = t.cast(Deployment, chosen)
    deployment.watch(bento_dir)


@click.group(name="deployment", cls=BentoMLCommandGroup)
def deployment_command():
    """Deployment Subcommands Groups"""


@deployment_command.command()
@shared_decorator()
@click.argument(
    "name",
    type=click.STRING,
    required=False,
)
@click.option(
    "--bento",
    type=click.STRING,
    help="Bento name or path to Bento project directory",
)
@click.option(
    "--access-authorization",
    type=click.BOOL,
    help="Enable access authorization",
)
@click.option(
    "--scaling-min",
    type=click.INT,
    help="Minimum scaling value",
)
@click.option(
    "--scaling-max",
    type=click.INT,
    help="Maximum scaling value",
)
@click.option(
    "--instance-type",
    type=click.STRING,
    help="Type of instance",
)
@click.option(
    "--strategy",
    type=click.Choice(
        [deployment_strategy.value for deployment_strategy in DeploymentStrategy]
    ),
    help="Deployment strategy",
)
@click.option(
    "--env",
    type=click.STRING,
    help="List of environment variables pass by --env key[=value] --env ...",
    multiple=True,
)
@click.option(
    "--secret",
    type=click.STRING,
    help="List of secret names pass by --secret name1, --secret name2, ...",
    multiple=True,
)
@click.option(
    "-f",
    "--config-file",
    type=click.File(),
    help="Configuration file path, mututally exclusive with other config options",
    default=None,
)
@click.option(
    "--config-dict",
    type=click.STRING,
    help="Configuration json string",
    default=None,
)
@inject
def update(  # type: ignore
    name: str | None,
    cluster: str | None,
    bento: str | None,
    access_authorization: bool | None,
    scaling_min: int | None,
    scaling_max: int | None,
    instance_type: str | None,
    strategy: str | None,
    env: tuple[str] | None,
    secret: tuple[str] | None,
    config_file: t.TextIO | None,
    config_dict: str | None,
    _cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
) -> None:
    """Update a deployment on BentoCloud.

    \b
    A deployment can be updated using parameters, or using config yaml file.
    You can also update bento by providing a project path or existing bento.
    """
    cfg_dict = None
    if config_dict is not None and config_dict != "":
        cfg_dict = json.loads(config_dict)
    config_params = DeploymentConfigParameters(
        name=name,
        bento=bento,
        cluster=cluster,
        access_authorization=access_authorization,
        scaling_max=scaling_max,
        scaling_min=scaling_min,
        instance_type=instance_type,
        strategy=strategy,
        envs=convert_env_to_dict(env),
        secrets=list(secret) if secret is not None else None,
        config_file=config_file,
        config_dict=cfg_dict,
        cli=True,
    )
    try:
        config_params.verify(create=False)
    except BentoMLException as e:
        raise_deployment_config_error(e, "update")
    deployment_info = _cloud_client.deployment.update(
        deployment_config_params=config_params
    )

    rich.print(f"Deployment [green]'{deployment_info.name}'[/] updated successfully.")


@deployment_command.command()
@click.argument(
    "bento",
    type=click.STRING,
    required=False,
)
@click.option(
    "-n",
    "--name",
    type=click.STRING,
    help="Deployment name",
)
@click.option(
    "--cluster",
    type=click.STRING,
    help="Name of the cluster",
)
@click.option(
    "--access-authorization",
    type=click.BOOL,
    help="Enable access authorization",
)
@click.option(
    "--scaling-min",
    type=click.INT,
    help="Minimum scaling value",
)
@click.option(
    "--scaling-max",
    type=click.INT,
    help="Maximum scaling value",
)
@click.option(
    "--instance-type",
    type=click.STRING,
    help="Type of instance",
)
@click.option(
    "--strategy",
    type=click.Choice(
        [deployment_strategy.value for deployment_strategy in DeploymentStrategy]
    ),
    help="Deployment strategy",
)
@click.option(
    "--env",
    type=click.STRING,
    help="List of environment variables pass by --env key[=value] --env ...",
    multiple=True,
)
@click.option(
    "--secret",
    type=click.STRING,
    help="List of secret names pass by --secret name1, --secret name2, ...",
    multiple=True,
)
@click.option(
    "-f",
    "--config-file",
    type=click.File(),
    help="Configuration file path",
    default=None,
)
@click.option(
    "-f",
    "--config-file",
    help="Configuration file path, mututally exclusive with other config options",
    default=None,
)
@click.option(
    "--config-dict",
    type=click.STRING,
    help="Configuration json string",
    default=None,
)
@inject
def apply(  # type: ignore
    bento: str | None,
    name: str | None,
    cluster: str | None,
    access_authorization: bool | None,
    scaling_min: int | None,
    scaling_max: int | None,
    instance_type: str | None,
    strategy: str | None,
    env: tuple[str] | None,
    secret: tuple[str] | None,
    config_file: str | t.TextIO | None,
    config_dict: str | None,
    _cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
) -> None:
    """Apply a deployment on BentoCloud.

    \b
    A deployment can be applied using config yaml file.
    """
    cfg_dict = None
    if config_dict is not None and config_dict != "":
        cfg_dict = json.loads(config_dict)
    config_params = DeploymentConfigParameters(
        name=name,
        bento=bento,
        cluster=cluster,
        access_authorization=access_authorization,
        scaling_max=scaling_max,
        scaling_min=scaling_min,
        instance_type=instance_type,
        strategy=strategy,
        envs=convert_env_to_dict(env),
        secrets=list(secret) if secret is not None else None,
        config_file=config_file,
        config_dict=cfg_dict,
        cli=True,
    )
    try:
        config_params.verify(create=False)
    except BentoMLException as e:
        raise_deployment_config_error(e, "apply")
    deployment_info = _cloud_client.deployment.apply(
        deployment_config_params=config_params
    )

    rich.print(f"Deployment [green]'{deployment_info.name}'[/] applied successfully.")


@deployment_command.command()
@click.argument(
    "bento",
    type=click.STRING,
    required=False,
)
@click.option(
    "-n",
    "--name",
    type=click.STRING,
    help="Deployment name",
)
@click.option(
    "--cluster",
    type=click.STRING,
    help="Name of the cluster",
)
@click.option(
    "--access-authorization",
    type=click.BOOL,
    help="Enable access authorization",
)
@click.option(
    "--scaling-min",
    type=click.INT,
    help="Minimum scaling value",
)
@click.option(
    "--scaling-max",
    type=click.INT,
    help="Maximum scaling value",
)
@click.option(
    "--instance-type",
    type=click.STRING,
    help="Type of instance",
)
@click.option(
    "--strategy",
    type=click.Choice(
        [deployment_strategy.value for deployment_strategy in DeploymentStrategy]
    ),
    help="Deployment strategy",
)
@click.option(
    "--env",
    type=click.STRING,
    help="List of environment variables pass by --env key[=value] --env ...",
    multiple=True,
)
@click.option(
    "--secret",
    type=click.STRING,
    help="List of secret names pass by --secret name1, --secret name2, ...",
    multiple=True,
)
@click.option(
    "-f",
    "--config-file",
    type=click.File(),
    help="Configuration file path",
    default=None,
)
@click.option(
    "--config-dict",
    type=click.STRING,
    help="Configuration json string",
    default=None,
)
@click.option(
    "--wait/--no-wait",
    type=click.BOOL,
    is_flag=True,
    help="Do not wait for deployment to be ready",
    default=True,
)
@click.option(
    "--timeout",
    type=click.INT,
    default=3600,
    help="Timeout for deployment to be ready in seconds",
)
def create(
    bento: str | None,
    name: str | None,
    cluster: str | None,
    access_authorization: bool | None,
    scaling_min: int | None,
    scaling_max: int | None,
    instance_type: str | None,
    strategy: str | None,
    env: tuple[str] | None,
    secret: tuple[str] | None,
    config_file: str | t.TextIO | None,
    config_dict: str | None,
    wait: bool,
    timeout: int,
) -> None:
    """Create a deployment on BentoCloud.

    \b
    Create a deployment using parameters, or using config yaml file.
    """
    create_deployment(
        bento=bento,
        name=name,
        cluster=cluster,
        access_authorization=access_authorization,
        scaling_min=scaling_min,
        scaling_max=scaling_max,
        instance_type=instance_type,
        strategy=strategy,
        env=env,
        secret=secret,
        config_file=config_file,
        config_dict=config_dict,
        wait=wait,
        timeout=timeout,
    )


@deployment_command.command()
@shared_decorator
@click.argument(
    "name",
    type=click.STRING,
    required=True,
)
@output_option
def get(  # type: ignore
    name: str,
    cluster: str | None,
    output: t.Literal["json", "default"],
) -> None:
    """Get a deployment on BentoCloud."""
    d = bentoml.deployment.get(name, cluster=cluster)
    if output == "json":
        info = json.dumps(d.to_dict(), indent=2, default=str)
        rich.print_json(info)
    else:
        info = yaml.dump(d.to_dict(), indent=2, sort_keys=False)
        rich.print(Syntax(info, "yaml", background_color="default"))


@deployment_command.command()
@shared_decorator
@click.argument(
    "name",
    type=click.STRING,
    required=True,
)
@click.option("--wait", is_flag=True, help="Wait for the deployment to be terminated")
def terminate(  # type: ignore
    name: str, cluster: str | None, wait: bool
) -> None:
    """Terminate a deployment on BentoCloud."""
    bentoml.deployment.terminate(name, cluster=cluster, wait=wait)
    rich.print(f"Deployment [green]'{name}'[/] terminated successfully.")


@deployment_command.command()
@click.argument(
    "name",
    type=click.STRING,
    required=True,
)
@shared_decorator
def delete(  # type: ignore
    name: str, cluster: str | None
) -> None:
    """Delete a deployment on BentoCloud."""
    bentoml.deployment.delete(name, cluster=cluster)
    rich.print(f"Deployment [green]'{name}'[/] deleted successfully.")


@deployment_command.command(name="list")
@click.option("--cluster", type=click.STRING, default=None, help="Name of the cluster.")
@click.option(
    "--search", type=click.STRING, default=None, help="Search for list request."
)
@click.option(
    "-o",
    "--output",
    help="Display the output of this command.",
    type=click.Choice(["json", "yaml", "table"]),
    default="table",
)
@click.option(
    "--label",
    "labels",
    type=click.STRING,
    multiple=True,
    default=None,
    help="Filter deployments by label(s).",
    metavar="KEY=VALUE",
)
def list_command(  # type: ignore
    cluster: str | None,
    search: str | None,
    labels: tuple[str, ...] | None,
    output: t.Literal["json", "yaml", "table"],
) -> None:
    """List existing deployments on BentoCloud."""
    if labels is not None:
        # For labels like ["env=prod", "team=infra"]
        # This will output: "label:env=prod label:team=infra"
        labels_query = " ".join(f"label:{label}" for label in labels)

    try:
        d_list = bentoml.deployment.list(cluster=cluster, search=search, q=labels_query)
    except BentoMLException as e:
        raise_deployment_config_error(e, "list")
    res: list[dict[str, t.Any]] = [d.to_dict() for d in d_list]
    if output == "table":
        table = Table(box=None, expand=True)
        table.add_column("Deployment", overflow="fold")
        table.add_column("created_at", overflow="fold")
        table.add_column("Bento", overflow="fold")
        table.add_column("Status", overflow="fold")
        table.add_column("Region", overflow="fold")
        for info in d_list:
            table.add_row(
                info.name,
                info.created_at,
                info.get_bento(refetch=False),
                info.get_status(refetch=False).status,
                info.cluster,
            )
        rich.print(table)
    elif output == "json":
        info = json.dumps(res, indent=2, default=str)
        rich.print_json(info)
    else:
        info = yaml.dump(res, indent=2, sort_keys=False)
        rich.print(Syntax(info, "yaml", background_color="default"))


@deployment_command.command()
@click.option("--cluster", type=click.STRING, default=None, help="Name of the cluster.")
@click.option(
    "-o",
    "--output",
    help="Display the output of this command.",
    type=click.Choice(["json", "yaml", "table"]),
    default="table",
)
@inject
def list_instance_types(  # type: ignore
    cluster: str | None,
    output: t.Literal["json", "yaml", "table"],
    _cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
) -> None:
    """List existing instance types in cluster on BentoCloud."""
    try:
        d_list = _cloud_client.deployment.list_instance_types(cluster=cluster)
    except BentoMLException as e:
        raise_deployment_config_error(e, "list_instance_types")
    res: list[dict[str, t.Any]] = [d.to_dict() for d in d_list]
    if output == "table":
        table = Table(box=None, expand=True)
        table.add_column("Name", overflow="fold")
        table.add_column("Price", overflow="fold")
        table.add_column("CPU", overflow="fold")
        table.add_column("Memory", overflow="fold")
        table.add_column("GPU", overflow="fold")
        table.add_column("GPU Type", overflow="fold")
        for info in d_list:
            table.add_row(
                info.name,
                info.price,
                info.cpu,
                info.memory,
                info.gpu,
                info.gpu_type,
            )
        rich.print(table)
    elif output == "json":
        info = json.dumps(res, indent=2, default=str)
        rich.print_json(info)
    else:
        info = yaml.dump(res, indent=2, sort_keys=False)
        rich.print(Syntax(info, "yaml", background_color="default"))


@inject
def create_deployment(
    bento: str | None = None,
    name: str | None = None,
    cluster: str | None = None,
    access_authorization: bool | None = None,
    scaling_min: int | None = None,
    scaling_max: int | None = None,
    instance_type: str | None = None,
    strategy: str | None = None,
    env: tuple[str] | None = None,
    secret: tuple[str] | None = None,
    config_file: str | t.TextIO | None = None,
    config_dict: str | None = None,
    wait: bool = True,
    timeout: int = 3600,
    dev: bool = False,
    _cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
) -> Deployment:
    cfg_dict = None
    if config_dict is not None and config_dict != "":
        cfg_dict = json.loads(config_dict)

    config_params = DeploymentConfigParameters(
        name=name,
        bento=bento,
        cluster=cluster,
        access_authorization=access_authorization,
        scaling_max=scaling_max,
        scaling_min=scaling_min,
        instance_type=instance_type,
        strategy=strategy,
        envs=convert_env_to_dict(env),
        secrets=list(secret) if secret is not None else None,
        config_file=config_file,
        config_dict=cfg_dict,
        cli=True,
        dev=dev,
    )

    try:
        config_params.verify()
    except BentoMLException as e:
        raise_deployment_config_error(e, "create")

    console = Console(highlight=False)
    with Spinner(console=console) as spinner:
        spinner.update("Creating deployment on BentoCloud")
        deployment = _cloud_client.deployment.create(
            deployment_config_params=config_params
        )
        spinner.log(
            f':white_check_mark: Created deployment "{deployment.name}" in cluster "{deployment.cluster}"'
        )
        spinner.log(f":laptop_computer: View Dashboard: {deployment.admin_console}")
        if wait:
            spinner.update(
                "[bold blue]Waiting for deployment to be ready, you can use --no-wait to skip this process[/]",
            )
            retcode = deployment.wait_until_ready(timeout=timeout, spinner=spinner)
            if retcode != 0:
                raise SystemExit(retcode)
        return deployment



from __future__ import annotations

import asyncio
import typing as t

import attrs
from simple_di import Provide
from simple_di import inject

from bentoml._internal.configuration.containers import BentoMLContainer
from bentoml.exceptions import BentoMLException

from .factory import Service

if t.TYPE_CHECKING:
    from _bentoml_impl.client.proxy import RemoteProxy

T = t.TypeVar("T")


_dependencies: list[Dependency[t.Any]] = []


async def cleanup() -> None:
    tasks = [dep.close() for dep in _dependencies]
    await asyncio.gather(*tasks)
    _dependencies.clear()


@attrs.define
class Dependency(t.Generic[T]):
    on: Service[T] | None = None
    deployment: str | None = None
    cluster: str | None = None
    url: str | None = None
    _resolved: t.Any = attrs.field(default=None, init=False)

    @t.overload
    def get(self: Dependency[None]) -> RemoteProxy[t.Any]: ...

    @t.overload
    def get(self: Dependency[T]) -> T: ...

    @inject
    def get(
        self: Dependency[T],
        *,
        runner_mapping: dict[str, str] = Provide[
            BentoMLContainer.remote_runner_mapping
        ],
    ) -> T | RemoteProxy[t.Any]:
        from _bentoml_impl.client.proxy import RemoteProxy

        media_type = "application/json"
        if self.deployment and self.url:
            raise BentoMLException("Cannot specify both deployment and url")
        if self.deployment:
            client = BentoMLContainer.rest_api_client.get()
            deployment = client.v2.get_deployment(self.deployment, self.cluster)
            try:
                self.url = deployment.urls[0]
            except IndexError:
                raise BentoMLException(
                    f"Deployment {self.deployment} does not have any URLs"
                )
        elif not self.url:
            if self.on is None:
                raise BentoMLException("Must specify one of on, deployment or url")
            if (key := self.on.name) in runner_mapping:
                self.url = runner_mapping[key]
                media_type = "application/vnd.bentoml+pickle"
            else:
                return self.on()

        return RemoteProxy(
            self.url, service=self.on, media_type=media_type
        ).as_service()

    @t.overload
    def __get__(self, instance: None, owner: t.Any) -> t.Self: ...

    @t.overload
    def __get__(
        self: Dependency[None], instance: t.Any, owner: t.Any
    ) -> RemoteProxy[t.Any]: ...

    @t.overload
    def __get__(self: Dependency[T], instance: t.Any, owner: t.Any) -> T: ...

    def __get__(
        self: Dependency[T], instance: t.Any, owner: t.Any
    ) -> Dependency[T] | RemoteProxy[t.Any] | T:
        if instance is None:
            return self
        if self._resolved is None:
            self._resolved = self.get()
            _dependencies.append(self)
        return self._resolved

    def __getattr__(self, name: str) -> t.Any:
        raise AttributeError("Dependency must be accessed as a class attribute")

    async def close(self) -> None:
        if self._resolved is None:
            return

        remote_proxy = t.cast("RemoteProxy[t.Any]", self._resolved)
        if asyncio.iscoroutinefunction(getattr(remote_proxy, "close", None)):
            await remote_proxy.close()


@t.overload
def depends(
    *,
    url: str | None = ...,
    deployment: str | None = ...,
    cluster: str | None = ...,
) -> Dependency[None]: ...


@t.overload
def depends(
    on: Service[T],
    *,
    url: str | None = ...,
    deployment: str | None = ...,
    cluster: str | None = ...,
) -> Dependency[T]: ...


def depends(
    on: Service[T] | None = None,
    *,
    url: str | None = None,
    deployment: str | None = None,
    cluster: str | None = None,
) -> Dependency[T]:
    """Create a dependency on other service or deployment

    Args:
        on: Service[T] | None: The service to depend on.
        url: str | None: The URL of the service to depend on.
        deployment: str | None: The deployment of the service to depend on.
        cluster: str | None: The cluster of the service to depend on.

    Examples:

    .. code-block:: python

        @bentoml.service
        class MyService:
            # depends on a service
            svc_a = bentoml.depends(SVC_A)
            # depends on a deployment
            svc_b = bentoml.depends(deployment="ci-iris")
            # depends on a remote service with url
            svc_c = bentoml.depends(url="http://192.168.1.1:3000")
            # For the latter two cases, the service can be given to provide more accurate types:
            svc_d = bentoml.depends(url="http://192.168.1.1:3000", on=SVC_D)
    """
    if on is not None and not isinstance(on, Service):
        raise TypeError("depends() expects a class decorated with @bentoml.service()")
    return Dependency(on, url=url, deployment=deployment, cluster=cluster)



# pylint: disable=unused-argument
from __future__ import annotations

import contextlib
import os
import tempfile
import typing as t
from typing import TYPE_CHECKING

import psutil
import pytest
from pytest import MonkeyPatch

import bentoml
from bentoml._internal.cloud import BentoCloudClient
from bentoml._internal.configuration import clean_bentoml_version
from bentoml._internal.configuration.containers import BentoMLContainer
from bentoml._internal.models import ModelContext
from bentoml._internal.utils.filesystem import validate_or_create_dir
from bentoml._internal.utils.lazy_loader import LazyLoader

if TYPE_CHECKING:
    import numpy as np
    from _pytest.config import Config
    from _pytest.config import ExitCode
    from _pytest.config.argparsing import Parser
    from _pytest.fixtures import FixtureRequest
    from _pytest.main import Session
    from _pytest.nodes import Item
    from _pytest.python import Metafunc

    from bentoml._internal.server.metrics.prometheus import PrometheusClient

else:
    np = LazyLoader("np", globals(), "numpy")


TEST_MODEL_CONTEXT = ModelContext(
    framework_name="testing",
    framework_versions={"testing": "v1"},
)

_RUN_GPU_TESTS_MARKER = "--run-gpu-tests"
_RUN_GRPC_TESTS_MARKER = "--run-grpc-tests"


@pytest.mark.tryfirst
def pytest_report_header(config: Config) -> list[str]:
    return [f"bentoml: version={clean_bentoml_version()}"]


@pytest.hookimpl
def pytest_addoption(parser: Parser) -> None:
    group = parser.getgroup("bentoml", "BentoML pytest plugins.")
    group.addoption(
        _RUN_GPU_TESTS_MARKER,
        action="store_true",
        default=False,
        help="run gpus related tests.",
    )
    group.addoption(
        _RUN_GRPC_TESTS_MARKER,
        action="store_true",
        default=False,
        help="run grpc related tests.",
    )


def pytest_configure(config: Config) -> None:
    # We will inject marker documentation here.
    config.addinivalue_line(
        "markers",
        "requires_gpus: requires GPU to run given test.",
    )
    config.addinivalue_line(
        "markers",
        "requires_grpc: requires gRPC support to run given test.",
    )


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item: Item) -> None:
    config = item.config
    if "requires_gpus" in item.keywords and not config.getoption(_RUN_GPU_TESTS_MARKER):
        item.add_marker(
            pytest.mark.skip(
                reason=f"need {_RUN_GPU_TESTS_MARKER} option to run gpus related tests."
            )
        )
    # We don't run gRPC tests on Windows
    if "requires_grpc" in item.keywords and not config.getoption(
        _RUN_GRPC_TESTS_MARKER
    ):
        item.add_marker(
            pytest.mark.skip(
                reason=f"need {_RUN_GRPC_TESTS_MARKER} option to run grpc related tests."
            )
        )


def _setup_deployment_mode(metafunc: Metafunc):
    """
    Setup deployment mode for test session.
    We will dynamically add this fixture to tests functions that has ``deployment_mode`` fixtures.

    Current matrix:
    - deployment_mode: ["container", "distributed", "standalone"]
    """
    if os.getenv("VSCODE_IPC_HOOK_CLI") and not os.getenv("GITHUB_CODESPACE_TOKEN"):
        # When running inside VSCode remote container locally, we don't have access to
        # exposed reserved ports, so we can't run container-based tests. However on GitHub
        # Codespaces, we can run container-based tests.
        # Note that inside the remote container, it is already running as a Linux container.
        deployment_mode = ["distributed", "standalone"]
    else:
        if os.environ.get("GITHUB_ACTIONS") and (psutil.WINDOWS or psutil.MACOS):
            # Due to GitHub Actions' limitation, we can't run container-based tests
            # on Windows and macOS. However, we can still running those tests on
            # local development.
            if psutil.MACOS:
                deployment_mode = ["distributed", "standalone"]
            else:
                deployment_mode = ["standalone"]
        else:
            if psutil.WINDOWS:
                deployment_mode = ["standalone", "container"]
            else:
                deployment_mode = ["distributed", "standalone", "container"]
    metafunc.parametrize("deployment_mode", deployment_mode, scope="session")


def _setup_model_store(metafunc: Metafunc):
    """Setup dummy models for test session."""
    with bentoml.models._create(  # type: ignore
        "testmodel",
        module=__name__,
        signatures={},
        context=TEST_MODEL_CONTEXT,
    ):
        pass
    with bentoml.models._create(  # type: ignore
        "testmodel",
        module=__name__,
        signatures={},
        context=TEST_MODEL_CONTEXT,
    ):
        pass
    with bentoml.models._create(  # type: ignore
        "anothermodel",
        module=__name__,
        signatures={},
        context=TEST_MODEL_CONTEXT,
    ):
        pass

    metafunc.parametrize(
        "model_store", [BentoMLContainer.model_store.get()], scope="session"
    )


@pytest.mark.tryfirst
def pytest_generate_tests(metafunc: Metafunc):
    if "deployment_mode" in metafunc.fixturenames:
        _setup_deployment_mode(metafunc)
    if "model_store" in metafunc.fixturenames:
        _setup_model_store(metafunc)


def _setup_session_environment(
    mp: MonkeyPatch, o: Session | Config, *pairs: tuple[str, str]
):
    """Setup environment variable for test session."""
    for p in pairs:
        key, value = p
        _ENV_VAR = os.environ.get(key, None)
        if _ENV_VAR is not None:
            mp.setattr(o, f"_original_{key}", _ENV_VAR, raising=False)
        os.environ[key] = value


def _setup_test_directory() -> tuple[str, str]:
    # Ensure we setup correct home and prometheus_multiproc_dir folders.
    # For any given test session.
    bentoml_home = tempfile.mkdtemp("bentoml-pytest")
    bentos = os.path.join(bentoml_home, "bentos")
    models = os.path.join(bentoml_home, "models")
    tmp_bentos = os.path.join(bentoml_home, "tmp_bentos")
    multiproc_dir = os.path.join(bentoml_home, "prometheus_multiproc_dir")
    validate_or_create_dir(bentos, models, tmp_bentos, multiproc_dir)

    # We need to set the below value inside container due to
    # the fact that each value is a singleton, and will be cached.
    BentoMLContainer.bentoml_home.set(bentoml_home)
    BentoMLContainer.bento_store_dir.set(bentos)
    BentoMLContainer.model_store_dir.set(models)
    BentoMLContainer.tmp_bento_store_dir.set(tmp_bentos)
    BentoMLContainer.prometheus_multiproc_dir.set(multiproc_dir)
    BentoMLContainer.bentocloud_client.set(BentoCloudClient(""))
    return bentoml_home, multiproc_dir


@pytest.mark.tryfirst
def pytest_sessionstart(session: Session) -> None:
    """Create a temporary directory for the BentoML home directory, then monkey patch to config."""
    from bentoml._internal.utils import analytics

    # We need to clear analytics cache before running tests.
    analytics.usage_stats.do_not_track.cache_clear()
    analytics.usage_stats._usage_event_debugging.cache_clear()  # type: ignore (private warning)

    mp = MonkeyPatch()
    config = session.config
    config.add_cleanup(mp.undo)

    _PYTEST_BENTOML_HOME, _PYTEST_MULTIPROC_DIR = _setup_test_directory()

    # The evironment variable patch ensures that we will
    # always build bento using bentoml from source, use the correct
    # test bentoml home directory, and setup prometheus multiproc directory.
    _setup_session_environment(
        mp,
        session,
        ("PROMETHEUS_MULTIPROC_DIR", _PYTEST_MULTIPROC_DIR),
        ("BENTOML_BUNDLE_LOCAL_BUILD", "True"),
        ("__BENTOML_DEBUG_USAGE", "False"),
        ("BENTOML_DO_NOT_TRACK", "True"),
    )

    _setup_session_environment(mp, config, ("BENTOML_HOME", _PYTEST_BENTOML_HOME))


def _teardown_session_environment(o: Session | Config, *variables: str):
    """Restore environment variable to original value."""
    for variable in variables:
        if hasattr(o, f"_original_{variable}"):
            os.environ[variable] = getattr(o, f"_original_{variable}")
        else:
            os.environ.pop(variable, None)


@pytest.mark.tryfirst
def pytest_sessionfinish(session: Session, exitstatus: int | ExitCode) -> None:
    config = session.config

    _teardown_session_environment(
        session,
        "BENTOML_BUNDLE_LOCAL_BUILD",
        "PROMETHEUS_MULTIPROC_DIR",
        "__BENTOML_DEBUG_USAGE",
        "BENTOML_DO_NOT_TRACK",
    )
    _teardown_session_environment(config, "BENTOML_HOME")

    # reset home and prometheus_multiproc_dir to default
    BentoMLContainer.prometheus_multiproc_dir.reset()


@pytest.fixture(scope="session")
def bentoml_home() -> str:
    """
    Return the BentoML home directory for the test session.
    This directory is created via ``pytest_sessionstart``.
    """
    return BentoMLContainer.bentoml_home.get()


@pytest.fixture(scope="session", autouse=True)
def clean_context() -> t.Generator[contextlib.ExitStack, None, None]:
    """
    Create a ExitStack to cleanup contextmanager.
    This fixture is available to all tests.
    """
    stack = contextlib.ExitStack()
    yield stack
    stack.close()


@pytest.fixture()
def img_file(tmpdir: str) -> str:
    """Create a random image/bmp file."""
    from PIL.Image import fromarray

    img_file_ = tmpdir.join("test_img.bmp")
    img = fromarray(np.random.randint(255, size=(10, 10, 3)).astype("uint8"))
    img.save(str(img_file_))
    return str(img_file_)


@pytest.fixture()
def bin_file(tmpdir: str) -> str:
    """Create a random binary file."""
    bin_file_ = tmpdir.join("bin_file.bin")
    with open(bin_file_, "wb") as of:
        of.write("â".encode("gb18030"))
    return str(bin_file_)


@pytest.fixture(scope="module", name="prom_client")
def fixture_metrics_client() -> PrometheusClient:
    """This fixtures return a PrometheusClient instance that can be used for testing."""
    return BentoMLContainer.metrics_client.get()


@pytest.fixture(scope="function", name="change_test_dir")
def fixture_change_dir(request: FixtureRequest) -> t.Generator[None, None, None]:
    """A fixture to change given test directory to the directory of the current running test."""
    os.chdir(
        request.fspath.dirname,  # type: ignore (bad pytest stubs)
    )
    yield
    os.chdir(
        request.config.invocation_dir,  # type: ignore (bad pytest stubs)
    )


@pytest.fixture(scope="function", autouse=True)
def fixture_reset_config() -> t.Generator[None, None, None]:
    """Reset BentoML config to default."""
    before = BentoMLContainer.config.get()
    try:
        yield
    finally:
        BentoMLContainer.config.set(before)



from __future__ import annotations

import functools
import logging
import math
import os
import re
import typing as t
from abc import ABC
from abc import abstractmethod

import psutil

from ..exceptions import BentoMLConfigException

logger = logging.getLogger(__name__)

_RESOURCE_REGISTRY: dict[str, t.Type[Resource[t.Any]]] = {}

T = t.TypeVar("T")


def get_resource(
    resources: dict[str, t.Any], resource_kind: str, validate: bool = True
) -> t.Any:
    if resource_kind not in _RESOURCE_REGISTRY:
        raise BentoMLConfigException(f"Unknown resource kind '{resource_kind}'.")

    resource: t.Type[Resource[t.Any]] = _RESOURCE_REGISTRY[resource_kind]

    if resource_kind in resources:
        if resources[resource_kind] == "system":
            return resource.from_system()
        else:
            res = resource.from_spec(resources[resource_kind])
            if validate:
                resource.validate(res)
            return res
    else:
        return None


def system_resources() -> dict[str, t.Any]:
    res: dict[str, t.Any] = {}
    for resource_kind, resource in _RESOURCE_REGISTRY.items():
        res[resource_kind] = resource.from_system()
    return res


class Resource(t.Generic[T], ABC):
    def __init_subclass__(cls, *, resource_id: str):  # pylint: disable=arguments-differ
        _RESOURCE_REGISTRY[resource_id] = cls

    @classmethod
    @abstractmethod
    def from_spec(cls, spec: t.Any) -> T:
        """
        Get an instance of this resource from user input. For example, a CPU resource
        might parse "10m" and return a CPU resource with 0.01 CPUs.
        """

    @classmethod
    @abstractmethod
    def from_system(cls) -> T:
        """
        Infer resource value from the system.
        """

    @classmethod
    @abstractmethod
    def validate(cls, val: T):
        """
        Validate that the resources are available on the current system.
        """


class CpuResource(Resource[float], resource_id="cpu"):
    @classmethod
    def from_spec(cls, spec: t.Any) -> float:
        """
        Convert spec to CpuResource.

        spec can be a float, int or string.
        - 1.0 -> 1.0
        - 1 -> 1.0
        - "1" -> 1.0
        - "10m" -> 0.01
        """
        if not isinstance(spec, (int, float, str)):
            raise TypeError("cpu must be int, float or str")

        if isinstance(spec, (int, float)):
            return float(spec)

        milli_match = re.match("([0-9]+)m", spec)
        if milli_match:
            return float(milli_match[1]) / 1000.0

        try:
            return float(spec)
        except ValueError:
            raise BentoMLConfigException(f"Invalid CPU resource limit '{spec}'. ")

    @classmethod
    def from_system(cls) -> float:
        if psutil.POSIX:
            return query_cgroup_cpu_count()
        else:
            return float(query_os_cpu_count())

    @classmethod
    def validate(cls, val: float):
        if val < 0:
            raise BentoMLConfigException(
                f"Invalid negative CPU resource limit '{val}'."
            )
        if not math.isclose(val, cls.from_system()) and val > cls.from_system():
            raise BentoMLConfigException(
                f"CPU resource limit {val} is greater than the system available: {cls.from_system()}"
            )


@functools.lru_cache(maxsize=1)
def query_cgroup_cpu_count() -> float:
    # Query active cpu processor count using cgroup v1 API, based on OpenJDK
    # implementation for `active_processor_count` using cgroup v1:
    # https://github.com/openjdk/jdk/blob/master/src/hotspot/os/linux/cgroupSubsystem_linux.cpp
    # For cgroup v2, see:
    # https://github.com/openjdk/jdk/blob/master/src/hotspot/os/linux/cgroupV2Subsystem_linux.cpp
    # Possible supports: cpuset.cpus on kubernetes
    def _read_cgroup_file(filename: str) -> float:
        with open(filename, "r", encoding="utf-8") as f:
            return int(f.read().strip())

    cgroup_root = "/sys/fs/cgroup/"
    cfs_quota_us_file = os.path.join(cgroup_root, "cpu", "cpu.cfs_quota_us")
    cfs_period_us_file = os.path.join(cgroup_root, "cpu", "cpu.cfs_period_us")
    cpu_max_file = os.path.join(cgroup_root, "cpu.max")

    quota = None

    if os.path.exists(cfs_quota_us_file) and os.path.exists(cfs_period_us_file):
        try:
            quota = _read_cgroup_file(cfs_quota_us_file) / _read_cgroup_file(
                cfs_period_us_file
            )
        except FileNotFoundError as err:
            logger.warning("Caught exception while calculating CPU quota: %s", err)
    # reading from cpu.max for cgroup v2
    elif os.path.exists(cpu_max_file):
        try:
            with open(cpu_max_file, "r", encoding="utf-8") as max_file:
                cfs_string = max_file.read()
                quota_str, period_str = cfs_string.split()
                if quota_str.isnumeric() and period_str.isnumeric():
                    quota = float(quota_str) / float(period_str)
                else:
                    # quota_str is "max" meaning the cpu quota is unset
                    quota = None
        except FileNotFoundError as err:
            logger.warning("Caught exception while calculating CPU quota: %s", err)
    if quota is not None and quota < 0:
        quota = None
    elif quota == 0:
        quota = 1

    os_cpu_count = float(os.cpu_count() or 1.0)

    limit_count = math.inf

    if quota:
        limit_count = quota

    return float(min(limit_count, os_cpu_count))


@functools.lru_cache(maxsize=1)
def query_os_cpu_count() -> int:
    cpu_count = os.cpu_count()
    if cpu_count is not None:
        return cpu_count
    logger.warning("Failed to determine CPU count, using 1 as default.")
    return 1


# class MemResource(Resource, resource_id="mem"):
#     @classmethod
#     def from_spec(cls, spec: t.Any):
#         assert isinstance(mem, (int, str)), "mem must be int or str"
#
#         if isinstance(mem, int):
#             return mem
#
#         unit_match = re.match("([0-9]+)([A-Za-z]{1,2})", mem)
#         mem_multipliers = {
#             "k": 1000,
#             "M": 1000**2,
#             "G": 1000**3,
#             "T": 1000**4,
#             "P": 1000**5,
#             "E": 1000**6,
#             "Ki": 1024,
#             "Mi": 1024**2,
#             "Gi": 1024**3,
#             "Ti": 1024**4,
#             "Pi": 1024**5,
#             "Ei": 1024**6,
#         }
#         if unit_match:
#             base = int(unit_match[1])
#             unit = unit_match[2]
#             if unit in mem_multipliers:
#                 return base * mem_multipliers[unit]
#         raise ValueError(f"Invalid MEM resource limit '{mem}'")


class NvidiaGpuResource(Resource[t.List[int]], resource_id="nvidia.com/gpu"):
    @classmethod
    def from_spec(cls, spec: int | str | list[int | str]) -> list[int]:
        if not isinstance(spec, (int, str, list)):
            raise TypeError(
                "NVidia GPU device IDs must be int, str or a list specifing the exact GPUs to use."
            )

        try:
            if isinstance(spec, int):
                if spec < -1:
                    raise ValueError
                return list(range(spec))
            elif isinstance(spec, str):
                return cls.from_spec(int(spec))
            else:
                return [int(x) for x in spec]
        except ValueError:
            raise BentoMLConfigException(
                f"Invalid NVidia GPU resource limit '{spec}'. "
            )

    @classmethod
    @functools.lru_cache(maxsize=1)
    def from_system(cls) -> list[int]:
        """
        query nvidia gpu count, available on Windows and Linux
        """
        import pynvml  # type: ignore

        try:
            pynvml.nvmlInit()
            device_count = pynvml.nvmlDeviceGetCount()
            return list(range(device_count))
        except (
            pynvml.NVMLError_LibraryNotFound,
            pynvml.NVMLError_DriverNotLoaded,
            OSError,
        ):
            logger.debug("GPU not detected. Unable to initialize pynvml lib.")
            return []
        finally:
            try:
                pynvml.nvmlShutdown()
            except Exception:  # pylint: disable=broad-except
                pass

    @classmethod
    def validate(cls, val: t.List[int]):
        if any([gpu_index < 0 for gpu_index in val]):
            raise BentoMLConfigException(f"Negative GPU device in {val}.")
        if any([gpu_index >= len(cls.from_system()) for gpu_index in val]):
            raise BentoMLConfigException(
                f"GPU device index in {val} is greater than the system available: {cls.from_system()}"
            )


def get_gpu_memory(dev: int) -> t.Tuple[float, float]:
    """
    Return Total Memory and Free Memory in given GPU device. in MiB
    """
    import pynvml.nvml  # type: ignore
    from pynvml.smi import nvidia_smi  # type: ignore

    unit_multiplier = {
        "PiB": 1024.0 * 1024 * 1024,
        "TiB": 1024.0 * 1024,
        "GiB": 1024.0,
        "MiB": 1.0,
        "KiB": 1.0 / 1024,
        "B": 1.0 / 1024 / 1024,
    }

    try:
        inst = nvidia_smi.getInstance()
        query: t.Dict[str, int] = inst.DeviceQuery(dev)  # type: ignore
    except (pynvml.nvml.NVMLError, OSError):
        return 0.0, 0.0

    try:
        gpus: t.List[t.Dict[str, t.Any]] = query.get("gpu", [])  # type: ignore
        gpu = gpus[dev]
        unit = gpu["fb_memory_usage"]["unit"]
        total = gpu["fb_memory_usage"]["total"] * unit_multiplier[unit]
        free = gpu["fb_memory_usage"]["free"] * unit_multiplier[unit]
        return total, free
    except IndexError:
        raise ValueError(f"Invalid GPU device index {dev}")
    except KeyError:
        raise RuntimeError(f"unexpected nvml query result: {query}")



from __future__ import annotations

import logging
import os
import sys
import typing as t

import click
import rich

if t.TYPE_CHECKING:
    P = t.ParamSpec("P")
    F = t.Callable[P, t.Any]


logger = logging.getLogger(__name__)

DEFAULT_DEV_SERVER_HOST = "127.0.0.1"


def deprecated_option(*param_decls: str, **attrs: t.Any):
    """Marks a given options as deprecated, and omit a warning when it's used"""
    deprecated = attrs.pop("deprecated", True)
    new_behaviour = attrs.pop("current_behaviour", None)
    assert new_behaviour is not None, "current_behaviour is required"

    def show_deprecated_callback(
        ctx: click.Context, param: click.Parameter, value: t.Any
    ):
        if value is not param.default and deprecated:
            name = "'--%(name)s'" if attrs.get("is_flag", False) else "'%(name)s'"
            DEPRECATION_WARNING = f"[yellow]DeprecationWarning: The parameter {name} is deprecated and will be removed in the future. (Current behaviour: %(new_behaviour)s)[/]"
            rich.print(
                DEPRECATION_WARNING
                % {"name": param.name, "new_behaviour": new_behaviour},
                file=sys.stderr,
            )

    def decorator(f: F[t.Any]) -> t.Callable[[F[t.Any]], click.Command]:
        msg = attrs.pop("help", "")
        msg += " (Deprecated)" if msg else "(Deprecated)"
        attrs.setdefault("help", msg)
        attrs.setdefault("callback", show_deprecated_callback)
        return click.option(*param_decls, **attrs)(f)

    return decorator


def build_serve_command() -> click.Group:
    from bentoml._internal.configuration.containers import BentoMLContainer
    from bentoml._internal.log import configure_server_logging
    from bentoml.grpc.utils import LATEST_PROTOCOL_VERSION
    from bentoml_cli.env_manager import env_manager
    from bentoml_cli.utils import AliasCommand
    from bentoml_cli.utils import BentoMLCommandGroup

    @click.group(name="serve", cls=BentoMLCommandGroup)
    def cli():
        pass

    @cli.command(aliases=["serve-http"], cls=AliasCommand)
    @click.argument("bento", type=click.STRING, default=".")
    @click.option(
        "--development",
        type=click.BOOL,
        help="Run the BentoServer in development mode",
        is_flag=True,
        default=False,
        show_default=True,
    )
    @deprecated_option(
        "--production",
        type=click.BOOL,
        help="Run BentoServer in production mode",
        current_behaviour="This is enabled by default. To run in development mode, use '--development'.",
        is_flag=True,
        default=True,
        show_default=False,
    )
    @click.option(
        "-p",
        "--port",
        type=click.INT,
        help="The port to listen on for the REST api server",
        envvar="BENTOML_PORT",
        show_envvar=True,
    )
    @click.option(
        "--host",
        type=click.STRING,
        help="The host to bind for the REST api server",
        envvar="BENTOML_HOST",
        show_envvar=True,
    )
    @click.option(
        "--api-workers",
        type=click.INT,
        help="Specify the number of API server workers to start. Default to number of available CPU cores in production mode",
        envvar="BENTOML_API_WORKERS",
        show_envvar=True,
        hidden=True,
    )
    @click.option(
        "--timeout",
        type=click.INT,
        help="Specify the timeout (seconds) for API server and runners",
        envvar="BENTOML_TIMEOUT",
        hidden=True,
    )
    @click.option(
        "--backlog",
        type=click.INT,
        help="The maximum number of pending connections.",
        show_default=True,
        hidden=True,
    )
    @click.option(
        "--reload",
        type=click.BOOL,
        is_flag=True,
        help="Reload Service when code changes detected",
        default=False,
        show_default=True,
    )
    @click.option(
        "--working-dir",
        type=click.Path(),
        help="When loading from source code, specify the directory to find the Service instance",
        default=None,
        show_default=True,
    )
    @click.option(
        "--ssl-certfile",
        type=str,
        help="SSL certificate file",
        show_default=True,
        hidden=True,
    )
    @click.option(
        "--ssl-keyfile",
        type=str,
        help="SSL key file",
        show_default=True,
        hidden=True,
    )
    @click.option(
        "--ssl-keyfile-password",
        type=str,
        help="SSL keyfile password",
        show_default=True,
        hidden=True,
    )
    @click.option(
        "--ssl-version",
        type=int,
        help="SSL version to use (see stdlib 'ssl' module)",
        show_default=True,
        hidden=True,
    )
    @click.option(
        "--ssl-cert-reqs",
        type=int,
        help="Whether client certificate is required (see stdlib 'ssl' module)",
        show_default=True,
        hidden=True,
    )
    @click.option(
        "--ssl-ca-certs",
        type=str,
        help="CA certificates file",
        show_default=True,
        hidden=True,
    )
    @click.option(
        "--ssl-ciphers",
        type=str,
        help="Ciphers to use (see stdlib 'ssl' module)",
        show_default=True,
        hidden=True,
    )
    @click.option(
        "--timeout-keep-alive",
        type=int,
        help="Close Keep-Alive connections if no new data is received within this timeout.",
        hidden=True,
    )
    @click.option(
        "--timeout-graceful-shutdown",
        type=int,
        default=None,
        help="Maximum number of seconds to wait for graceful shutdown. After this timeout, the server will start terminating requests.",
        show_default=True,
        hidden=True,
    )
    @env_manager
    def serve(  # type: ignore (unused warning)
        bento: str,
        development: bool,
        port: int,
        host: str,
        api_workers: int,
        timeout: int | None,
        backlog: int,
        reload: bool,
        working_dir: str | None,
        ssl_certfile: str | None,
        ssl_keyfile: str | None,
        ssl_keyfile_password: str | None,
        ssl_version: int | None,
        ssl_cert_reqs: int | None,
        ssl_ca_certs: str | None,
        ssl_ciphers: str | None,
        timeout_keep_alive: int | None,
        timeout_graceful_shutdown: int | None,
        **attrs: t.Any,
    ) -> None:
        """Start a HTTP BentoServer from a given 🍱

        \b
        BENTO is the serving target, it can be the import as:
        - the import path of a 'bentoml.Service' instance
        - a tag to a Bento in local Bento store
        - a folder containing a valid 'bentofile.yaml' build file with a 'service' field, which provides the import path of a 'bentoml.Service' instance
        - a path to a built Bento (for internal & debug use only)

        e.g.:

        \b
        Serve from a bentoml.Service instance source code (for development use only):
            'bentoml serve fraud_detector.py:svc'

        \b
        Serve from a Bento built in local store:
            'bentoml serve fraud_detector:4tht2icroji6zput3suqi5nl2'
            'bentoml serve fraud_detector:latest'

        \b
        Serve from a Bento directory:
            'bentoml serve ./fraud_detector_bento'

        \b
        If '--reload' is provided, BentoML will detect code and model store changes during development, and restarts the service automatically.

        \b
        The '--reload' flag will:
        - be default, all file changes under '--working-dir' (default to current directory) will trigger a restart
        - when specified, respect 'include' and 'exclude' under 'bentofile.yaml' as well as the '.bentoignore' file in '--working-dir', for code and file changes
        - all model store changes will also trigger a restart (new model saved or existing model removed)
        """
        from bentoml import Service
        from bentoml._internal.service.loader import load

        configure_server_logging()
        if working_dir is None:
            if os.path.isdir(os.path.expanduser(bento)):
                working_dir = os.path.expanduser(bento)
            else:
                working_dir = "."
        if sys.path[0] != working_dir:
            sys.path.insert(0, working_dir)
        svc = load(bento_identifier=bento, working_dir=working_dir)
        if isinstance(svc, Service):
            # bentoml<1.2
            from bentoml.serving import serve_http_production

            if development:
                serve_http_production(
                    bento,
                    working_dir=working_dir,
                    port=port,
                    host=DEFAULT_DEV_SERVER_HOST if not host else host,
                    backlog=backlog,
                    api_workers=1,
                    timeout=timeout,
                    ssl_keyfile=ssl_keyfile,
                    ssl_certfile=ssl_certfile,
                    ssl_keyfile_password=ssl_keyfile_password,
                    ssl_version=ssl_version,
                    ssl_cert_reqs=ssl_cert_reqs,
                    ssl_ca_certs=ssl_ca_certs,
                    ssl_ciphers=ssl_ciphers,
                    reload=reload,
                    development_mode=True,
                    timeout_keep_alive=timeout_keep_alive,
                    timeout_graceful_shutdown=timeout_graceful_shutdown,
                )
            else:
                serve_http_production(
                    bento,
                    working_dir=working_dir,
                    port=port,
                    host=host,
                    api_workers=api_workers,
                    timeout=timeout,
                    ssl_keyfile=ssl_keyfile,
                    ssl_certfile=ssl_certfile,
                    ssl_keyfile_password=ssl_keyfile_password,
                    ssl_version=ssl_version,
                    ssl_cert_reqs=ssl_cert_reqs,
                    ssl_ca_certs=ssl_ca_certs,
                    ssl_ciphers=ssl_ciphers,
                    reload=reload,
                    development_mode=False,
                    timeout_keep_alive=timeout_keep_alive,
                    timeout_graceful_shutdown=timeout_graceful_shutdown,
                )
        else:
            # bentoml>=1.2

            from _bentoml_impl.server import serve_http

            svc.inject_config()
            serve_http(
                bento,
                working_dir=working_dir,
                host=host,
                port=port,
                backlog=backlog,
                timeout=timeout,
                ssl_certfile=ssl_certfile,
                ssl_keyfile=ssl_keyfile,
                ssl_keyfile_password=ssl_keyfile_password,
                ssl_version=ssl_version,
                ssl_cert_reqs=ssl_cert_reqs,
                ssl_ca_certs=ssl_ca_certs,
                ssl_ciphers=ssl_ciphers,
                development_mode=development,
                reload=reload,
                timeout_keep_alive=timeout_keep_alive,
                timeout_graceful_shutdown=timeout_graceful_shutdown,
            )

    @cli.command(name="serve-grpc", hidden=True)
    @click.argument("bento", type=click.STRING, default=".")
    @click.option(
        "--development",
        type=click.BOOL,
        help="Run the BentoServer in development mode",
        is_flag=True,
        default=False,
        show_default=True,
    )
    @deprecated_option(
        "--production",
        type=click.BOOL,
        help="Run BentoServer in production mode",
        current_behaviour="This is enabled by default. To run in development mode, use '--development'.",
        is_flag=True,
        default=True,
        show_default=False,
    )
    @click.option(
        "-p",
        "--port",
        type=click.INT,
        default=BentoMLContainer.grpc.port.get(),
        help="The port to listen on for the REST api server",
        envvar="BENTOML_PORT",
        show_default=True,
    )
    @click.option(
        "--host",
        type=click.STRING,
        default=BentoMLContainer.grpc.host.get(),
        help="The host to bind for the gRPC server",
        envvar="BENTOML_HOST",
        show_default=True,
    )
    @click.option(
        "--api-workers",
        type=click.INT,
        default=BentoMLContainer.api_server_workers.get(),
        help="Specify the number of API server workers to start. Default to number of available CPU cores in production mode",
        envvar="BENTOML_API_WORKERS",
        show_default=True,
    )
    @click.option(
        "--reload",
        type=click.BOOL,
        is_flag=True,
        help="Reload Service when code changes detected",
        default=False,
        show_default=True,
    )
    @click.option(
        "--backlog",
        type=click.INT,
        default=BentoMLContainer.api_server_config.backlog.get(),
        help="The maximum number of pending connections.",
        show_default=True,
    )
    @click.option(
        "--working-dir",
        type=click.Path(),
        help="When loading from source code, specify the directory to find the Service instance",
        default=None,
        show_default=True,
    )
    @click.option(
        "--enable-reflection",
        is_flag=True,
        default=BentoMLContainer.grpc.reflection.enabled.get(),
        type=click.BOOL,
        help="Enable reflection.",
        show_default=True,
    )
    @click.option(
        "--enable-channelz",
        is_flag=True,
        default=BentoMLContainer.grpc.channelz.enabled.get(),
        type=click.BOOL,
        help="Enable Channelz. See https://github.com/grpc/proposal/blob/master/A14-channelz.md.",
    )
    @click.option(
        "--max-concurrent-streams",
        default=BentoMLContainer.grpc.max_concurrent_streams.get(),
        type=click.INT,
        help="Maximum number of concurrent incoming streams to allow on a http2 connection.",
        show_default=True,
    )
    @click.option(
        "--ssl-certfile",
        type=str,
        default=BentoMLContainer.ssl.certfile.get(),
        help="SSL certificate file",
        show_default=True,
    )
    @click.option(
        "--ssl-keyfile",
        type=str,
        default=BentoMLContainer.ssl.keyfile.get(),
        help="SSL key file",
        show_default=True,
    )
    @click.option(
        "--ssl-ca-certs",
        type=str,
        default=BentoMLContainer.ssl.ca_certs.get(),
        help="CA certificates file",
        show_default=True,
    )
    @click.option(
        "-pv",
        "--protocol-version",
        type=click.Choice(["v1", "v1alpha1"]),
        help="Determine the version of generated gRPC stubs to use.",
        default=LATEST_PROTOCOL_VERSION,
        show_default=True,
    )
    def serve_grpc(  # type: ignore (unused warning)
        bento: str,
        development: bool,
        port: int,
        host: str,
        api_workers: int | None,
        backlog: int,
        reload: bool,
        working_dir: str | None,
        ssl_certfile: str | None,
        ssl_keyfile: str | None,
        ssl_ca_certs: str | None,
        enable_reflection: bool,
        enable_channelz: bool,
        max_concurrent_streams: int | None,
        protocol_version: str,
        **attrs: t.Any,
    ):
        """Start a gRPC BentoServer from a given 🍱

        \b
        BENTO is the serving target, it can be the import as:
        - the import path of a 'bentoml.Service' instance
        - a tag to a Bento in local Bento store
        - a folder containing a valid 'bentofile.yaml' build file with a 'service' field, which provides the import path of a 'bentoml.Service' instance
        - a path to a built Bento (for internal & debug use only)

        e.g.:

        \b
        Serve from a bentoml.Service instance source code (for development use only):
            'bentoml serve-grpc fraud_detector.py:svc'

        \b
        Serve from a Bento built in local store:
            'bentoml serve-grpc fraud_detector:4tht2icroji6zput3suqi5nl2'
            'bentoml serve-grpc fraud_detector:latest'

        \b
        Serve from a Bento directory:
            'bentoml serve-grpc ./fraud_detector_bento'

        If '--reload' is provided, BentoML will detect code and model store changes during development, and restarts the service automatically.

        \b
        The '--reload' flag will:
        - be default, all file changes under '--working-dir' (default to current directory) will trigger a restart
        - when specified, respect 'include' and 'exclude' under 'bentofile.yaml' as well as the '.bentoignore' file in '--working-dir', for code and file changes
        - all model store changes will also trigger a restart (new model saved or existing model removed)
        """
        configure_server_logging()
        if working_dir is None:
            if os.path.isdir(os.path.expanduser(bento)):
                working_dir = os.path.expanduser(bento)
            else:
                working_dir = "."

        from bentoml.serving import serve_grpc_production

        if development:
            serve_grpc_production(
                bento,
                working_dir=working_dir,
                port=port,
                host=DEFAULT_DEV_SERVER_HOST if not host else host,
                backlog=backlog,
                api_workers=1,
                ssl_keyfile=ssl_keyfile,
                ssl_certfile=ssl_certfile,
                ssl_ca_certs=ssl_ca_certs,
                max_concurrent_streams=max_concurrent_streams,
                reflection=enable_reflection,
                channelz=enable_channelz,
                protocol_version=protocol_version,
                reload=reload,
                development_mode=True,
            )
        else:
            serve_grpc_production(
                bento,
                working_dir=working_dir,
                port=port,
                backlog=backlog,
                host=host,
                api_workers=api_workers,
                ssl_keyfile=ssl_keyfile,
                ssl_certfile=ssl_certfile,
                ssl_ca_certs=ssl_ca_certs,
                max_concurrent_streams=max_concurrent_streams,
                reflection=enable_reflection,
                channelz=enable_channelz,
                protocol_version=protocol_version,
                reload=reload,
                development_mode=False,
            )

    return cli


serve_command = build_serve_command()



from __future__ import annotations

import json
import os
import typing as t

import click
import rich
import yaml
from rich.syntax import Syntax
from rich.table import Table
from simple_di import Provide
from simple_di import inject

from bentoml import Tag
from bentoml._internal.bento.bento import DEFAULT_BENTO_BUILD_FILES
from bentoml._internal.bento.build_config import BentoBuildConfig
from bentoml._internal.container import BentoMLContainer
from bentoml._internal.utils import human_readable_size
from bentoml._internal.utils.filesystem import calc_dir_size
from bentoml._internal.utils.filesystem import resolve_user_filepath
from bentoml.exceptions import InvalidArgument
from bentoml.models import import_model
from bentoml_cli.utils import BentoMLCommandGroup
from bentoml_cli.utils import is_valid_bento_name
from bentoml_cli.utils import is_valid_bento_tag

if t.TYPE_CHECKING:
    from click import Context
    from click import Parameter

    from bentoml._internal.bento import BentoStore
    from bentoml._internal.cloud import BentoCloudClient
    from bentoml._internal.models import ModelStore


def parse_delete_targets_argument_callback(
    ctx: Context,
    params: Parameter,
    value: t.Any,  # pylint: disable=unused-argument
) -> t.Any:
    if value is None:
        return value
    value = " ".join(value)
    if "," in value:
        delete_targets = value.split(",")
    else:
        delete_targets = value.split()
    delete_targets = list(map(str.strip, delete_targets))
    for delete_target in delete_targets:
        if not (
            is_valid_bento_tag(delete_target) or is_valid_bento_name(delete_target)
        ):
            raise click.BadParameter(
                'Bad formatting. Please present a valid bento bundle name or "name:version" tag. For list of bento bundles, separate delete targets by ",", for example: "my_service:v1,my_service:v2,classifier"'
            )
    return delete_targets


@click.group(name="models", cls=BentoMLCommandGroup)
def model_command():
    """Model Subcommands Groups"""


@model_command.command()
@click.argument("model_tag", type=click.STRING)
@click.option(
    "-o",
    "--output",
    type=click.Choice(["json", "yaml", "path"]),
    default="yaml",
)
@inject
def get(
    model_tag: str,
    output: str,
    model_store: ModelStore = Provide[BentoMLContainer.model_store],
) -> None:  # type: ignore (not accessed)
    """Print Model details by providing the model_tag

    \b
    bentoml get iris_clf:qojf5xauugwqtgxi
    bentoml get iris_clf:qojf5xauugwqtgxi --output=json
    """
    model = model_store.get(model_tag)

    if output == "path":
        rich.print(model.path)
    elif output == "json":
        info = json.dumps(model.info.to_dict(), indent=2, default=str)
        rich.print_json(info)
    else:
        rich.print(Syntax(str(model.info.dump()), "yaml", background_color="default"))


@model_command.command(name="list")
@click.argument("model_name", type=click.STRING, required=False)
@click.option(
    "-o",
    "--output",
    type=click.Choice(["json", "yaml", "table"]),
    default="table",
)
@inject
def list_models(
    model_name: str,
    output: str,
    model_store: ModelStore = Provide[BentoMLContainer.model_store],
) -> None:  # type: ignore (not accessed)
    """List Models in local store

    \b
    # show all models saved
    $ bentoml models list

    \b
    # show all versions of bento with the name FraudDetector
    $ bentoml models list FraudDetector
    """

    models = model_store.list(model_name)
    res = [
        {
            "tag": str(model.tag),
            "module": model.info.module,
            "size": human_readable_size(calc_dir_size(model.path)),
            "creation_time": model.info.creation_time.astimezone().strftime(
                "%Y-%m-%d %H:%M:%S"
            ),
        }
        for model in sorted(models, key=lambda x: x.info.creation_time, reverse=True)
    ]
    if output == "json":
        info = json.dumps(res, indent=2)
        rich.print_json(info)
    elif output == "yaml":
        info = yaml.safe_dump(res, indent=2)
        rich.print(Syntax(info, "yaml"))
    else:
        table = Table(box=None)
        table.add_column("Tag")
        table.add_column("Module")
        table.add_column("Size")
        table.add_column("Creation Time")
        for model in res:
            table.add_row(
                model["tag"],
                model["module"],
                model["size"],
                model["creation_time"],
            )
        rich.print(table)


@model_command.command()
@click.argument(
    "delete_targets",
    nargs=-1,
    callback=parse_delete_targets_argument_callback,
    required=True,
)
@click.option(
    "-y",
    "--yes",
    "--assume-yes",
    is_flag=True,
    help="Skip confirmation when deleting a specific model",
)
@inject
def delete(
    delete_targets: str,
    yes: bool,
    model_store: ModelStore = Provide[BentoMLContainer.model_store],
    bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
) -> None:  # type: ignore (not accessed)
    """Delete Model in local model store.

    \b
    Examples:
        * Delete single model by "name:version", e.g: `bentoml models delete iris_clf:v1`
        * Bulk delete all models with a specific name, e.g.: `bentoml models delete iris_clf`
        * Bulk delete multiple models by name and version, separated by ",", e.g.: `bentoml models delete iris_clf:v1,iris_clf:v2`
        * Bulk delete multiple models by name and version, separated by " ", e.g.: `bentoml models delete iris_clf:v1 iris_clf:v2`
        * Bulk delete without confirmation, e.g.: `bentoml models delete IrisClassifier --yes`
    """  # noqa
    from bentoml.exceptions import BentoMLException

    def check_model_is_used(tag: Tag) -> None:
        in_use: list[Tag] = []
        for bento in bento_store.list():
            if bento._model_store is not None:
                continue
            if any(model.tag == tag for model in bento.info.all_models):
                in_use.append(bento.tag)
        if in_use:
            raise BentoMLException(
                f"Model {tag} is being used by the following bentos and can't be deleted:\n  "
                + "\n  ".join(map(str, in_use))
            )

    def delete_target(target: str) -> None:
        tag = Tag.from_str(target)

        if tag.version is None:
            to_delete_models = model_store.list(target)
        else:
            to_delete_models = [model_store.get(tag)]

        for model in to_delete_models:
            if yes:
                delete_confirmed = True
            else:
                delete_confirmed = click.confirm(f"delete model {model.tag}?")

            if delete_confirmed:
                check_model_is_used(model.tag)
                model_store.delete(model.tag)
                rich.print(f"[green]{model}[/] deleted.")

    for target in delete_targets:
        delete_target(target)


@model_command.command()
@click.argument("model_tag", type=click.STRING)
@click.argument("out_path", type=click.STRING, default="", required=False)
@inject
def export(
    model_tag: str,
    out_path: str,
    model_store: ModelStore = Provide[BentoMLContainer.model_store],
) -> None:  # type: ignore (not accessed)
    """Export a Model to an external archive file

    arguments:

    \b
    MODEL_TAG: model identifier
    OUT_PATH: output path of exported model.
        If this argument is not provided, model is exported to name-version.bentomodel in the current directory.
        Besides native .bentomodel format, we also support formats like tar('tar'), tar.gz ('gz'), tar.xz ('xz'), tar.bz2 ('bz2'), and zip.

    examples:

    \b
    bentoml models export FraudDetector:latest
    bentoml models export FraudDetector:latest ./my_model.bentomodel
    bentoml models export FraudDetector:20210709_DE14C9 ./my_model.bentomodel
    bentoml models export FraudDetector:20210709_DE14C9 s3://mybucket/models/my_model.bentomodel
    """
    bentomodel = model_store.get(model_tag)
    out_path = bentomodel.export(out_path)
    rich.print(f"[green]{bentomodel}[/] exported to {out_path}.")


@model_command.command(name="import")
@click.argument("model_path", type=click.STRING)
def import_from(model_path: str) -> None:  # type: ignore (not accessed)
    """Import a previously exported Model archive file

    bentoml models import ./my_model.bentomodel
    bentoml models import s3://mybucket/models/my_model.bentomodel
    """
    bentomodel = import_model(model_path)
    rich.print(f"[green]{bentomodel}[/] imported.")


@model_command.command()
@click.argument("model_tag", type=click.STRING, required=False)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force pull from remote model store to local and overwrite even if it already exists in local",
)
@click.option(
    "-F",
    "--bentofile",
    help="Path to bentofile. Default to 'bentofile.yaml'",
)
@click.pass_context
@inject
def pull(
    ctx: click.Context,
    model_tag: str | None,
    force: bool,
    bentofile: str | None,
    cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
):
    """Pull Model from a remote model store. If model_tag is not provided,
    it will pull models defined in bentofile.yaml.
    """
    from click.core import ParameterSource

    if model_tag is not None:
        if ctx.get_parameter_source("bentofile") != ParameterSource.DEFAULT:
            rich.print("-f bentofile is ignored when model_tag is provided")
        cloud_client.model.pull(model_tag, force=force)
        return

    if bentofile:
        try:
            bentofile = resolve_user_filepath(bentofile, None)
        except FileNotFoundError:
            raise InvalidArgument(f'file "{bentofile}" not found')
        else:
            build_config = BentoBuildConfig.from_file(bentofile)
    else:
        for filename in DEFAULT_BENTO_BUILD_FILES:
            if os.path.exists(filename):
                build_config = BentoBuildConfig.from_file(filename)
                break
        else:
            raise InvalidArgument(
                "No bentofile.yaml or pyproject.toml found, please provide a bentofile path"
            )

    if not build_config.models:
        raise InvalidArgument(
            "No model to pull, please provide a model tag or define models in bentofile.yaml"
        )
    for model_spec in build_config.models:
        cloud_client.model.pull(
            model_spec.tag,
            force=force,
            query=model_spec.filter,
        )


@model_command.command()
@click.argument("model_tag", type=click.STRING)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Forced push to remote model store even if it exists in remote",
)
@click.option(
    "-t",
    "--threads",
    default=10,
    help="Number of threads to use for upload",
)
@inject
def push(
    model_tag: str,
    force: bool,
    threads: int,
    model_store: ModelStore = Provide[BentoMLContainer.model_store],
    cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
):  # type: ignore (not accessed)
    """Push Model to a remote model store."""
    from _bentoml_sdk.models import BentoModel

    model_obj = BentoModel(model_tag)
    if model_obj.stored is None:
        raise click.ClickException(f"Model {model_tag} not found in local store")
    cloud_client.model.push(
        model_obj,
        force=force,
        threads=threads,
    )



from __future__ import annotations

import attrs

from bentoml._internal.cloud.client import RestApiClient

from .base import Spinner
from .bento import BentoAPI
from .config import DEFAULT_ENDPOINT
from .config import CloudClientConfig
from .deployment import DeploymentAPI
from .model import ModelAPI
from .secret import SecretAPI
from .yatai import YataiClient as YataiClient


@attrs.frozen
class BentoCloudClient:
    """
    BentoCloudClient is a client for the BentoCloud API.

    Args:
        api_key: The API key to use for the client. env: BENTO_CLOUD_API_KEY
        endpoint: The endpoint to use for the client. env: BENTO_CLOUD_ENDPOINT
        timeout: The timeout to use for the client. Defaults to 60 seconds.

    Attributes:
        bento: Bento API
        model: Model API
        deployment: Deployment API
        secret: Secret API
    """

    bento: BentoAPI
    model: ModelAPI
    deployment: DeploymentAPI
    secret: SecretAPI

    def __init__(
        self,
        api_key: str | None = None,
        endpoint: str = DEFAULT_ENDPOINT,
        timeout: int = 60,
    ) -> None:
        if api_key is None:
            from ..configuration.containers import BentoMLContainer

            cfg = CloudClientConfig.get_config()
            ctx = cfg.get_context(BentoMLContainer.cloud_context.get())
            api_key = ctx.api_token
            endpoint = ctx.endpoint

        client = RestApiClient(endpoint, api_key, timeout)
        spinner = Spinner()
        bento = BentoAPI(client, spinner=spinner)
        model = ModelAPI(client, spinner=spinner)
        deployment = DeploymentAPI(client)
        secret = SecretAPI(client)

        self.__attrs_init__(
            bento=bento, model=model, deployment=deployment, secret=secret
        )

    @classmethod
    def for_context(cls, context: str | None = None) -> "BentoCloudClient":
        cfg = CloudClientConfig.get_config()
        ctx = cfg.get_context(context)
        return cls(ctx.api_token, ctx.endpoint)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: bentoml/grpc/v1alpha1/service.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#bentoml/grpc/v1alpha1/service.proto\x12\x15\x62\x65ntoml.grpc.v1alpha1\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1egoogle/protobuf/wrappers.proto\"\xa3\x03\n\x07Request\x12\x10\n\x08\x61pi_name\x18\x01 \x01(\t\x12\x31\n\x07ndarray\x18\x03 \x01(\x0b\x32\x1e.bentoml.grpc.v1alpha1.NDArrayH\x00\x12\x35\n\tdataframe\x18\x05 \x01(\x0b\x32 .bentoml.grpc.v1alpha1.DataFrameH\x00\x12/\n\x06series\x18\x06 \x01(\x0b\x32\x1d.bentoml.grpc.v1alpha1.SeriesH\x00\x12+\n\x04\x66ile\x18\x07 \x01(\x0b\x32\x1b.bentoml.grpc.v1alpha1.FileH\x00\x12,\n\x04text\x18\x08 \x01(\x0b\x32\x1c.google.protobuf.StringValueH\x00\x12&\n\x04json\x18\t \x01(\x0b\x32\x16.google.protobuf.ValueH\x00\x12\x35\n\tmultipart\x18\n \x01(\x0b\x32 .bentoml.grpc.v1alpha1.MultipartH\x00\x12\x1a\n\x10serialized_bytes\x18\x02 \x01(\x0cH\x00\x42\t\n\x07\x63ontentJ\x04\x08\x04\x10\x05J\x04\x08\x0b\x10\x0e\"\x92\x03\n\x08Response\x12\x31\n\x07ndarray\x18\x01 \x01(\x0b\x32\x1e.bentoml.grpc.v1alpha1.NDArrayH\x00\x12\x35\n\tdataframe\x18\x03 \x01(\x0b\x32 .bentoml.grpc.v1alpha1.DataFrameH\x00\x12/\n\x06series\x18\x05 \x01(\x0b\x32\x1d.bentoml.grpc.v1alpha1.SeriesH\x00\x12+\n\x04\x66ile\x18\x06 \x01(\x0b\x32\x1b.bentoml.grpc.v1alpha1.FileH\x00\x12,\n\x04text\x18\x07 \x01(\x0b\x32\x1c.google.protobuf.StringValueH\x00\x12&\n\x04json\x18\x08 \x01(\x0b\x32\x16.google.protobuf.ValueH\x00\x12\x35\n\tmultipart\x18\t \x01(\x0b\x32 .bentoml.grpc.v1alpha1.MultipartH\x00\x12\x1a\n\x10serialized_bytes\x18\x02 \x01(\x0cH\x00\x42\t\n\x07\x63ontentJ\x04\x08\x04\x10\x05J\x04\x08\n\x10\x0e\"\xde\x02\n\x04Part\x12\x31\n\x07ndarray\x18\x01 \x01(\x0b\x32\x1e.bentoml.grpc.v1alpha1.NDArrayH\x00\x12\x35\n\tdataframe\x18\x03 \x01(\x0b\x32 .bentoml.grpc.v1alpha1.DataFrameH\x00\x12/\n\x06series\x18\x05 \x01(\x0b\x32\x1d.bentoml.grpc.v1alpha1.SeriesH\x00\x12+\n\x04\x66ile\x18\x06 \x01(\x0b\x32\x1b.bentoml.grpc.v1alpha1.FileH\x00\x12,\n\x04text\x18\x07 \x01(\x0b\x32\x1c.google.protobuf.StringValueH\x00\x12&\n\x04json\x18\x08 \x01(\x0b\x32\x16.google.protobuf.ValueH\x00\x12\x1a\n\x10serialized_bytes\x18\x04 \x01(\x0cH\x00\x42\x10\n\x0erepresentationJ\x04\x08\x02\x10\x03J\x04\x08\t\x10\x0e\"\x95\x01\n\tMultipart\x12<\n\x06\x66ields\x18\x01 \x03(\x0b\x32,.bentoml.grpc.v1alpha1.Multipart.FieldsEntry\x1aJ\n\x0b\x46ieldsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12*\n\x05value\x18\x02 \x01(\x0b\x32\x1b.bentoml.grpc.v1alpha1.Part:\x02\x38\x01\"\xf1\x02\n\x04\x46ile\x12\x37\n\x04kind\x18\x01 \x01(\x0e\x32$.bentoml.grpc.v1alpha1.File.FileTypeH\x00\x88\x01\x01\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\x0c\"\x95\x02\n\x08\x46ileType\x12\x19\n\x15\x46ILE_TYPE_UNSPECIFIED\x10\x00\x12\x11\n\rFILE_TYPE_CSV\x10\x01\x12\x17\n\x13\x46ILE_TYPE_PLAINTEXT\x10\x02\x12\x12\n\x0e\x46ILE_TYPE_JSON\x10\x03\x12\x13\n\x0f\x46ILE_TYPE_BYTES\x10\x04\x12\x11\n\rFILE_TYPE_PDF\x10\x05\x12\x11\n\rFILE_TYPE_PNG\x10\x06\x12\x12\n\x0e\x46ILE_TYPE_JPEG\x10\x07\x12\x11\n\rFILE_TYPE_GIF\x10\x08\x12\x11\n\rFILE_TYPE_BMP\x10\t\x12\x12\n\x0e\x46ILE_TYPE_TIFF\x10\n\x12\x12\n\x0e\x46ILE_TYPE_WEBP\x10\x0b\x12\x11\n\rFILE_TYPE_SVG\x10\x0c\x42\x07\n\x05_kind\"Q\n\tDataFrame\x12\x14\n\x0c\x63olumn_names\x18\x01 \x03(\t\x12.\n\x07\x63olumns\x18\x02 \x03(\x0b\x32\x1d.bentoml.grpc.v1alpha1.Series\"\xa1\x01\n\x06Series\x12\x17\n\x0b\x62ool_values\x18\x01 \x03(\x08\x42\x02\x10\x01\x12\x18\n\x0c\x66loat_values\x18\x02 \x03(\x02\x42\x02\x10\x01\x12\x18\n\x0cint32_values\x18\x03 \x03(\x05\x42\x02\x10\x01\x12\x18\n\x0cint64_values\x18\x06 \x03(\x03\x42\x02\x10\x01\x12\x15\n\rstring_values\x18\x05 \x03(\t\x12\x19\n\rdouble_values\x18\x04 \x03(\x01\x42\x02\x10\x01\"\xc8\x03\n\x07NDArray\x12\x33\n\x05\x64type\x18\x01 \x01(\x0e\x32$.bentoml.grpc.v1alpha1.NDArray.DType\x12\r\n\x05shape\x18\x02 \x03(\x05\x12\x15\n\rstring_values\x18\x05 \x03(\t\x12\x18\n\x0c\x66loat_values\x18\x03 \x03(\x02\x42\x02\x10\x01\x12\x19\n\rdouble_values\x18\x04 \x03(\x01\x42\x02\x10\x01\x12\x17\n\x0b\x62ool_values\x18\x06 \x03(\x08\x42\x02\x10\x01\x12\x18\n\x0cint32_values\x18\x07 \x03(\x05\x42\x02\x10\x01\x12\x18\n\x0cint64_values\x18\x08 \x03(\x03\x42\x02\x10\x01\x12\x19\n\ruint32_values\x18\t \x03(\rB\x02\x10\x01\x12\x19\n\ruint64_values\x18\n \x03(\x04\x42\x02\x10\x01\"\xa9\x01\n\x05\x44Type\x12\x15\n\x11\x44TYPE_UNSPECIFIED\x10\x00\x12\x0f\n\x0b\x44TYPE_FLOAT\x10\x01\x12\x10\n\x0c\x44TYPE_DOUBLE\x10\x02\x12\x0e\n\nDTYPE_BOOL\x10\x03\x12\x0f\n\x0b\x44TYPE_INT32\x10\x04\x12\x0f\n\x0b\x44TYPE_INT64\x10\x05\x12\x10\n\x0c\x44TYPE_UINT32\x10\x06\x12\x10\n\x0c\x44TYPE_UINT64\x10\x07\x12\x10\n\x0c\x44TYPE_STRING\x10\x08\x32Y\n\x0c\x42\x65ntoService\x12I\n\x04\x43\x61ll\x12\x1e.bentoml.grpc.v1alpha1.Request\x1a\x1f.bentoml.grpc.v1alpha1.Response\"\x00\x42i\n\x19\x63om.bentoml.grpc.v1alpha1B\x0cServiceProtoP\x01Z0github.com/bentoml/bentoml/grpc/v1alpha1;service\x90\x01\x01\xf8\x01\x01\xa2\x02\x03SVCb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bentoml.grpc.v1alpha1.service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\031com.bentoml.grpc.v1alpha1B\014ServiceProtoP\001Z0github.com/bentoml/bentoml/grpc/v1alpha1;service\220\001\001\370\001\001\242\002\003SVC'
  _globals['_MULTIPART_FIELDSENTRY']._options = None
  _globals['_MULTIPART_FIELDSENTRY']._serialized_options = b'8\001'
  _globals['_SERIES'].fields_by_name['bool_values']._options = None
  _globals['_SERIES'].fields_by_name['bool_values']._serialized_options = b'\020\001'
  _globals['_SERIES'].fields_by_name['float_values']._options = None
  _globals['_SERIES'].fields_by_name['float_values']._serialized_options = b'\020\001'
  _globals['_SERIES'].fields_by_name['int32_values']._options = None
  _globals['_SERIES'].fields_by_name['int32_values']._serialized_options = b'\020\001'
  _globals['_SERIES'].fields_by_name['int64_values']._options = None
  _globals['_SERIES'].fields_by_name['int64_values']._serialized_options = b'\020\001'
  _globals['_SERIES'].fields_by_name['double_values']._options = None
  _globals['_SERIES'].fields_by_name['double_values']._serialized_options = b'\020\001'
  _globals['_NDARRAY'].fields_by_name['float_values']._options = None
  _globals['_NDARRAY'].fields_by_name['float_values']._serialized_options = b'\020\001'
  _globals['_NDARRAY'].fields_by_name['double_values']._options = None
  _globals['_NDARRAY'].fields_by_name['double_values']._serialized_options = b'\020\001'
  _globals['_NDARRAY'].fields_by_name['bool_values']._options = None
  _globals['_NDARRAY'].fields_by_name['bool_values']._serialized_options = b'\020\001'
  _globals['_NDARRAY'].fields_by_name['int32_values']._options = None
  _globals['_NDARRAY'].fields_by_name['int32_values']._serialized_options = b'\020\001'
  _globals['_NDARRAY'].fields_by_name['int64_values']._options = None
  _globals['_NDARRAY'].fields_by_name['int64_values']._serialized_options = b'\020\001'
  _globals['_NDARRAY'].fields_by_name['uint32_values']._options = None
  _globals['_NDARRAY'].fields_by_name['uint32_values']._serialized_options = b'\020\001'
  _globals['_NDARRAY'].fields_by_name['uint64_values']._options = None
  _globals['_NDARRAY'].fields_by_name['uint64_values']._serialized_options = b'\020\001'
  _globals['_REQUEST']._serialized_start=125
  _globals['_REQUEST']._serialized_end=544
  _globals['_RESPONSE']._serialized_start=547
  _globals['_RESPONSE']._serialized_end=949
  _globals['_PART']._serialized_start=952
  _globals['_PART']._serialized_end=1302
  _globals['_MULTIPART']._serialized_start=1305
  _globals['_MULTIPART']._serialized_end=1454
  _globals['_MULTIPART_FIELDSENTRY']._serialized_start=1380
  _globals['_MULTIPART_FIELDSENTRY']._serialized_end=1454
  _globals['_FILE']._serialized_start=1457
  _globals['_FILE']._serialized_end=1826
  _globals['_FILE_FILETYPE']._serialized_start=1540
  _globals['_FILE_FILETYPE']._serialized_end=1817
  _globals['_DATAFRAME']._serialized_start=1828
  _globals['_DATAFRAME']._serialized_end=1909
  _globals['_SERIES']._serialized_start=1912
  _globals['_SERIES']._serialized_end=2073
  _globals['_NDARRAY']._serialized_start=2076
  _globals['_NDARRAY']._serialized_end=2532
  _globals['_NDARRAY_DTYPE']._serialized_start=2363
  _globals['_NDARRAY_DTYPE']._serialized_end=2532
  _globals['_BENTOSERVICE']._serialized_start=2534
  _globals['_BENTOSERVICE']._serialized_end=2623
_builder.BuildServices(DESCRIPTOR, 'bentoml.grpc.v1alpha1.service_pb2', _globals)
# @@protoc_insertion_point(module_scope)
